///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (bayes.hh)
///*

///guard
#ifndef HH_BAYES
#define HH_BAYES
///*

///header
/* class name:	Bayes
 * 
 * description:	parameter estimation class
 * 
 * author:		Christian Himpe
 * 
 * notes:		...
 * 
 * todo:        ...
 * 
 * version:		1.0
 */
///*

///includes
#include <omp.h>
#include "vector.hh"
#include "matrix.hh"
#include "structs.hh"
#include "system.hh"
#include "solver.hh"
#include "feed.hh"
#include "drift.hh"
#include "utils.hh"
///*

///definitions
class Bayes
{
	private:
		const syspara& sys;												//General Parameter Structure
		Drift& drift;													//Reference to Drift Class
		System& system;													//Reference to System Class
		Solver& solver;													//Reference to Solver Class
		Feed& feed;														//Reference to Feed Class
		vector y;														//Observed Datapoints
		vector nt;														//Parameter Prior Means
		matrix ct;														//Parameter Prior Covariances
		vector nty;														//Parameter Posterior Means
		matrix cty;														//Parameter Posterior Covariances
		vector nby;														//Parameter Posterior Drift
		
		int paras;														//Number of (system) Parameters
		int points;														//Number of (all) Datapoints
		
		matrix jacobian(const vector& h) const;										//Numerical Evaluation of Parameter Jacobian dh/dtheta
	public:
		Bayes(const syspara& s,System& x,Solver& v,Feed& f,Drift& d);	//Constructor
		~Bayes() { ; };													//Destructor
		void expmax();													//Expectation Maximization Algorithm
		double postprob(const vector& c,double t);						//Posterior Probabilities
};
///*

///implementation
matrix Bayes::jacobian(const vector& h) const
{
	matrix r(points,paras);
	vector q(paras,ROW);
	vector g(points,ROW);

	omp_set_num_threads(threads);
	#pragma omp parallel for firstprivate(g,q) shared(r)
	for(int i=0;i<paras;i++)
	{
		System x(sys,feed);
		Solver s(sys,x,drift);
		q = nty;
		q(i,q(i)+sys.bayes_finitediff);
		x.reset(q);
		g = (s.solve()-h)*(1.0/sys.bayes_finitediff);
		#pragma omp critical
		{ r.paste(g,0,i); }
	}
	return r;
}

Bayes::Bayes(const syspara& s,System& x,Solver& v,Feed& f,Drift& d)
 : sys(s),
   system(x),
   solver(v),
   feed(f),
   drift(d),
   y(s.states*s.steps,ROW),
   nt(x.getpara()),
   ct(x.getcova()),
   nty(x.getpara()),
   cty(x.getcova()),
   nby(d.getpara()),
   paras(x.getparas()),
   points(s.states*s.steps)
{
	for(int i=0;i<sys.steps;i++) { y.paste(feed.data(i),i*sys.states); }//Concat Datapoints To Super Column Vector
	tty("[    "); tty(paras); say("    ]");
}

void Bayes::expmax()
{
	int xparas = drift.getparas();										//Number of Drift Parameters
	if(sys.drift_order==-1) { xparas = 0; }								//Exclude Drift if requested by config
	int epoints = points + paras;										//Number of Extended Datapoints
	int eparas = paras + xparas;										//Drift extended parameters
	
	matrix u(paras,paras,1.0);
	matrixarray s(sys.states,matrix(sys.states,sys.states));

	vector h(points,ROW);
	vector nd(paras,ROW);												//extended posterior differences 
	vector ne(eparas,ROW);												//drift extended extended posterior differences
	vector res(epoints,ROW);											//residuals
	vector ye(epoints,ROW);												//extended residuals
	matrix x(drift.global());											//drift matrix with dim (points,xparas)
	matrix j(points,paras);												//parameter jacobian
	matrix je(epoints,eparas);											//extended parameter jacobian

	matrix cee(epoints,epoints);										//temporal extended covariances
	matrix jetcee(eparas,epoints);										// jet*cee
	matrix jetceet(epoints,eparas);										// jet*cee transposed
	matrix ctyjetcee(eparas,epoints);									// cty*jetcee
	matrix ctye(eparas,eparas);											// extended conditional covariance

	vector l(sys.states,ROW,0.001);										//hyperparameters lambda
	matrixarray q(sys.states,matrix(epoints,epoints));					//temporal correlation matrices
	matrix p(epoints,epoints);											//residual forming matrix
	vector py(epoints,ROW);												// p*y
	matrixarray qp(sys.states,matrix(epoints,epoints));					// q[i]*p
	matrixarray qc(sys.states,matrix(epoints,epoints));					// q[i]*cee
	matrixarray qj(sys.states,matrix(epoints,eparas));					// q[i]*jetceet
	matrixarray qk(sys.states,matrix(eparas,eparas));					// ctyjetcee*q[i]
	vector df(sys.states,ROW);
	matrix d2f(sys.states,sys.states);
	
	for(int i=0;i<sys.states;i++)
	{
		s[i](i,i,1.0);
		q[i].paste(matrix(sys.steps,sys.steps,1.0).kronecker(s[i]),0,0);
	}

	int cnt = 0;
	tty("[");
	say("");
	while(true && sys.bayes_minloops!=0)
	{
		///E-STEP
		h = solver.solve();												//solve system with current parameters

		j = jacobian(h);
		je.clear();
		je.paste(j,0,0);												//paste parameter jacobain to upper left of extended
		je.paste(u,points,0);											//paste unit matrix to lower left
		if(sys.drift_order!=-1) { je.paste(x,0,paras); }				//paste drift matrix to upper right
																		//lower right remains zero	
		
		res = (y-h);													//compute residuals																		
		ye.paste(res,0);												//assemble extended residual vector
		ye.paste(nt-nty,points);										//use priors as further residuals 
																
		cee.clear();
		for(int i=0;i<sys.states;i++) { cee.saxpy(q[i],/*exp*/(l(i))); }	//upper left of extended covariances is sum of all with l*q's
		cee.paste(ct,points,points);									//lower right of extended covariances is prior covariances
		cee = cee.diaginverse();										//invert cee

		jetceet = cee * je;												//this saves one transposition opposed to other way round
		jetcee = jetceet.transpose();									//and precalculates jetceet which is needed in M-Step
	
		ctye = (jetcee * je).choleskyinverse();							//compute posterior covariances		

		ctyjetcee = ctye * jetcee;										//precompute ctyjetcee
		ne = ctyjetcee * ye; 											//compute parameter differences 
			
		nd = ne.copy(0,paras);											//copy posterior parameters
		nby = ne.copy(paras,xparas);									//copy drift parameters
	
		nty += nd;														//set new posterior means by adding computed differences
		tty("|");
		///*

		///M-STEP
		//original M-Step 
		//~ p = cee - cee * je * cty * je.transpose() * cee;				//compute residual forming matrix
		//~ #pragma omp parallel for firstprivate(ye,p,q) shared(df,d2f)
		//~ for(int i=0;i<sys.states;i++)
		//~ {
			//~ df(i,-0.5*(p*q[i]).trace()+0.5*ye.transpose()*p.transpose()*q[i]*p*ye);	//compute components of gradient dF/dl
			//~ for(int j=0;j<sys.states;j++)
			//~ {
				//~ d2f(i,j,-0.5*((p*q[i])*(p*q[j])).trace());				//compute components of jacobian d2F/dl2
			//~ }
		//~ }
		//~ l -= d2f.choleskysolver(df);									//update hyperparameters l													
		//*
		
		//Improved M-Step
		py = cee * ye - jetceet * ne;
		#pragma omp parallel for firstprivate(q,cee,jetceet,ctyjetcee) shared(qc,qj,qk)
		for(int i=0;i<sys.states;i++)
		{
			qc[i] = q[i] * cee;
			qj[i] = q[i] * jetceet;
			qk[i] = ctyjetcee * qj[i];
		}
		#pragma omp parallel for firstprivate(py,qc,qk,qj,ctyjetcee) shared(df,d2f)
		for(int i=0;i<sys.states;i++)
		{
				df(i,py*(q[i]*py) - qc[i].trace() + qk[i].trace());
				for(int j=i;j<sys.states;j++)
				{
					d2f(i,j,qc[i].producttrace(qc[j])
					-(qc[j]*qj[i]).producttrace(ctyjetcee)
					-(qc[i]*qj[j]).producttrace(ctyjetcee)
					+qk[i].producttrace(qk[j]));
					d2f(j,i,d2f(i,j));
				}
		}
		l += d2f.choleskysolver(df);
		//*
		tty("|");
		///*
		
		system.reset(nty);												//reset system parameters with new posterior means
		drift.reset(nby);												//reset drift parameters
		if(cnt<10) { tty(" "); } tty(cnt); tty(".: ");  tty(res.length()); tty(" "); say(nd.length()); 
		if(nd.length()<sys.bayes_converge || (++cnt>=sys.bayes_minloops && sys.bayes_minloops!=-1) ) { tty("|"); break; }
	}
	say("]");
	
	cty = ctye.copy(0,0,paras,paras);
}

double Bayes::postprob(const vector& c,double t)
{
	return cumnorm((c*nty-t)/sqrt(c.transpose()*cty*c));
}
///*

#endif 
