///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (dyneeg.hh)
///*

///guard
#ifndef HH_DYNEEG
#define HH_DYNEEG
///*

///header
/* class name:	Dyneeg
 * 
 * description:	class for eeg dynamic system sub model
 * 
 * author:		Christian Himpe
 * 
 * notes:		...
 * 
 * todo:        ...
 * 
 * version:		1.0
 */
///*

///includes
#include <cmath>
#include "utils.hh"
#include "dynbase.hh"
#include "vector.hh"
#include "matrix.hh"
#include "structs.hh"
///*

///declarations
#define DEDIMPERSTATE 9													//Dynamic EEG Dimensions per State

#define DEHEPM 4.0														//Dynamic EEG Excitatory Synaptic Potential Prior Mean
#define DEHEPV 0.00625 													//Dynamic EEG Excitatory Synaptic Potential Prior Variance
#define DEHIPM 32.0														//Dynamic EEG Inhibitory Synaptic Potential Prior Mean
#define DEHIPV 0.00625													//Dynamic EEG Inhibitory Synaptic Potential Prior Variance
#define DETEPM 0.008													//Dynamic EEG Excitatory Lumped Rate Prior Mean
#define DETEPV 0.00625													//Dynamic EEG Excitatory Lumped Rate Prior Variance
#define DETIPM 0.016													//Dynamic EEG Inhibitory Lumped Rate Prior Mean
#define DETIPV 0.00625													//Dynamic EEG Inhibitory Lumped Rate Prior Variance
#define DEDDPM 0.002													//Dynamic EEG Auto Transmission Delay Prior Mean
#define DEDDPV 0.00625													//Dynamic EEG Auto Transmission Delay Prior Variance
#define DEDFPM 0.016													//Dynamic EEG Transmission Delay Prior Mean
#define DEDFPV 0.00625													//Dynamic EEG Transmission Delay Prior Variance
#define DET1PM 1.0														//Dynamic EEG Input Delay and Shape 1 Prior Mean
#define DET1PV 0.00625													//Dynamic EEG Input Delay and Shape 1 Prior Variance
#define DET2PM 16.0 													//Dynamic EEG Input Delay and Shape 2 Prior Mean
#define DET2PV 0.00625													//Dynamic EEG Input Delay and Shape 2 Prior Variance
#define DET3PM 0.0														//Dynamic EEG Input Discrete Cosine Prior Mean
#define DET3PV 1.0														//Dynamic EEG Input Discrete Cosine Prior Variance
#define DECFPM 32.0														//Dynamic EEG Forward Connection Prior Mean
#define DECFPV 0.5														//Dynamic EEG Forward Connection Prior Variance
#define DECBPM 16.0														//Dynamic EEG Backward Connection Prior Mean
#define DECBPV 0.5														//Dynamic EEG Backward Connection Prior Variance
#define DECLPM 4.0														//Dynamic EEG Lateral Connection Prior Mean
#define DECLPV 0.5														//Dynamic EEG Lateral Connection Prior Variance
#define DEGPM  1.0														//Dynamic EEG Connection Gain Prior Mean
#define DEGPV  0.5														//Dynamic EEG Connection Gain Prior Variance
#define DECUPM 1.0														//Dynamic EEG Forward Input Prior Mean
#define DECUPV 0.5 														//Dynamic EEG Forward Input Prior Variance

#define E0 0.5															//Sigmoid Parameter E0
#define E1 2.0*E0														//Sigmoid Parameter E1
#define R  0.56															//Sigmoid Parameter R
///*

///definitions
class Dyneeg : public Dynbase
{
	private:
		matrix cf;														//Forward Connectivity Matrix
		matrix cl;														//Latent Connectivity Matrix
		matrix cb;														//Backward Connectivity Matrix
		vector cu;														//Input Connectivity Vector
		matrixarray g; 													//Gain Matrices
		vector he;														//Excitatory Synaptic Potential
		vector te;														//Excitatory Lumped Rate
		vector hi;														//Inhibitory Synaptic Potential
		vector ti;														//Inhibitory Lumped Rate
		double dd;														//Auto Transmission Delay
		double df;														//Transmission Delay
		double t1;														//Input Shape and Delay 1
		double t2;														//Input Shape and Delay 2
		vector t3;														//Input Discrete Cosine
		
		double g1;														//Intrinsic Connection 1
		double g2;														//Intrinsic Connection 2
		double g3;														//Intrinsic Connection 3
		double g4;														//Intrinsic Connection 4

		vector inppa;													//precomputed input constants
		
		matrix herte;													// he/te
		matrix drte;													// 2/te
		matrix rtete;													// 1/(te*te)
		matrix hirti;													// hi/ti
		matrix drti;													// 2/ti
		matrix rtiti;													// 1/(ti*ti)		
		matrix cfcl;													// cf+cl
		matrix cbcl;													// cb+cl

		int diadelay;
		int offdelay;
		vectorque back;													//Backstates
		
		inline vector sigmoid(const vector& v) const;					//Combined Sigmoid

		inline double input(double t) const;							//Transform Input
		inline void setup();											//use helper variables for common computations
	public:
		Dyneeg(const syspara& s,void* d);								//Constructors
		~Dyneeg() { ; };												//Destructor
		vector initial();												//Initial Values
		vector righthand(const vector& z,const vector& u,double t) const;	//Righthandside
		void   reset(const vector& d);									//Reset Paramters with Given
		void   last(const vector& d);									//Set Previous Result as Backstep
		vector getpara(bool t) const;									//Get Parameters (overloaded from base)
		bool   isstable() const;										//Check if Linearised System is Stable
};
///*

///implementation
vector Dyneeg::sigmoid(const vector& v) const
{
	vector r(sys.states,ROW);
	for(int i=0;i<sys.states;++i) { r(i,( E1 / (1.0+exp(-R*v(i))) ) - E0 ); }
	return r;
}

double Dyneeg::input(double t) const
{
	double r = inppa(0)*pow(t,inppa(1))*exp(inppa(2)*t);
	for(int i=0;i<sys.dyneeg_flucsorder;++i) { r += t3(i)*cos(inppa(3+i)*t); }
	return r;
}

void Dyneeg::setup()
{
	inppa(0,pow(t2,t1)/gamma(t1));
	inppa(1,t1-1.0);
	inppa(2,-t2);
	for(int i=0;i<sys.dyneeg_flucsorder;++i) { inppa(3+i,2.0*M_PI*double(i)); }
	
	diadelay = iround(1000.0*dd)-1;
	offdelay = iround(1000.0*df)-1;

	for(int i=0;i<sys.states;i++)
	{
		herte(i,i,he(i)/te(i));
		drte(i,i,2.0/te(i));
		rtete(i,i,1.0/(te(i)*te(i)));
		hirti(i,i,hi(i)/ti(i));
		drti(i,i,2.0/ti(i));
		rtiti(i,i,1.0/(ti(i)*ti(i)));
	}

	cfcl = (cf+cl);
	cbcl = (cb+cl);
}

Dyneeg::Dyneeg(const syspara& s,void* d)
 : Dynbase(s,DEDIMPERSTATE,(3/*+s.events-1*/)*s.states*s.states-3*s.states+4*s.states+3+s.dyneeg_flucsorder+count(s.inpc.c_str(),'1')),
   cf(s.states,s.states),
   cl(s.states,s.states),
   cb(s.states,s.states),
   g(s.events,matrix(s.states,s.states,1.0,1.0)),
   cu(s.states,ROW),
   g1(s.dyneeg_intrinsic),
   g2(s.dyneeg_intrinsic*0.8),
   g3(s.dyneeg_intrinsic*0.25),
   g4(s.dyneeg_intrinsic*0.25),
   t3(s.dyneeg_flucsorder,ROW),
   cfcl(s.states,s.states),
   cbcl(s.states,s.states),
   he(s.states,ROW),
   te(s.states,ROW),
   hi(s.states,ROW),
   ti(s.states,ROW),
   herte(s.states,s.states,ROW),
   drte(s.states,s.states,ROW),
   rtete(s.states,s.states,ROW),
   hirti(s.states,s.states,ROW),
   drti(s.states,s.states,ROW),
   rtiti(s.states,s.states,ROW),
   inppa(3+sys.dyneeg_flucsorder,ROW)
{
	int curr = 0;														//current index of parameter vector
	
	if(sys.type==SIM)													//Preset Parameters for Simulations
	{
		dyneegpara* de = static_cast<dyneegpara*>(d);					//Cast Parameter Structure  
		for(int i=0;i<sys.states;i++) { he(i,de->he); }
		for(int i=0;i<sys.states;i++) { te(i,de->te); }
		for(int i=0;i<sys.states;i++) { hi(i,de->hi); }
		for(int i=0;i<sys.states;i++) { ti(i,de->ti); }
		dd=de->dd;
		df=de->df;
		t1=de->t1;
		t2=de->t2;
		for(int j=0;j<sys.dyneeg_flucsorder;j++) { t3(j,0.0); }
		for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++) { cf(j,k,de->cf[j*sys.states+k]); } }
		for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++) { cb(j,k,de->cb[j*sys.states+k]); } }
		for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++) { cl(j,k,de->cl[j*sys.states+k]); } }
		for(int j=0;j<sys.states;j++) { cu(j,de->cu[j]); }
		for(int i=1;i<sys.events;i++) { for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++,curr++) { g[i](j,k,de->g[j*sys.states+k]); } } }
	}
	else
	{
		for(int i=0;i<sys.states;i++) { he(i,DEHEPM); cova(curr,curr,DEHEPV); curr++; }
		for(int i=0;i<sys.states;i++) { te(i,DETEPM); cova(curr,curr,DETEPV); curr++; }
		for(int i=0;i<sys.states;i++) { hi(i,DEHIPM); cova(curr,curr,DEHIPV); curr++; }
		for(int i=0;i<sys.states;i++) { ti(i,DETIPM); cova(curr,curr,DETIPV); curr++; }
		dd=DEDDPM; cova(curr,curr,DEDDPV); curr++;
		df=DEDFPM; cova(curr,curr,DEDFPV); curr++;
		t1=DET1PM; //cova(curr,curr,DET1PV); curr++;
		t2=DET2PM; cova(curr,curr,DET2PV); curr++;
		for(int j=0;j<sys.dyneeg_flucsorder;j++) { t3(j,DET3PM); cova(curr,curr,DET3PV); curr++; }
		for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++) { if(j!=k) { cf(j,k,DECFPM); cova(curr,curr,DECFPV); curr++; } } }
		for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++) { if(j!=k) { cb(j,k,DECBPM); cova(curr,curr,DECBPV); curr++; } } }
		for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++) { if(j!=k) { cl(j,k,DECLPM); cova(curr,curr,DECLPV); curr++; } } }
		for(int j=0;j<sys.states;j++) { if(sys.inpc.c_str()[j]=='1') { cu(j,DECUPM); cova(curr,curr,DECUPV); curr++; } }
		//for(int i=1;i<sys.events;i++) { for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++,curr++) { g[i](j,k,DEGPM); cova(curr,curr,DEGPV); } } }
	}

	setup();
}

vector Dyneeg::initial()
{
	back.clear();
	for(int j=0;j<=offdelay;j++) { back.push_back(vector(DEDIMPERSTATE*sys.states,ROW)); }
	return vector(DEDIMPERSTATE*sys.states,ROW);
}

vector Dyneeg::righthand(const vector& z,const vector& u,double t) const
{
	vector r(DEDIMPERSTATE*sys.states,ROW); 
	vector so0(sigmoid(back[offdelay].copy(0,sys.states)));
	vector sd0(sigmoid(back[diadelay].copy(0,sys.states)));
	vector sd1(sigmoid(back[diadelay].copy(1*sys.states,sys.states)));
	vector sd7(sigmoid(back[diadelay].copy(7*sys.states,sys.states)));
	matrix a(g[int(u(0))]);
	matrix hertegii(herte.hadamard(a));
	double v = sign(u(0)) * input(t-u(1));

	vectorarray x(DEDIMPERSTATE,vector(sys.states,ROW)); 
	vectorarray y(DEDIMPERSTATE,vector(sys.states,ROW));
	for(int i=0;i<DEDIMPERSTATE;i++) { x[i] = z.copy(i*sys.states,sys.states); }
	y[0] = x[5] - x[6];
	y[1] = x[4];
	y[2] = x[5];
	y[7] = x[8];
	y[3] = x[6];
	y[4] = hertegii*( cfcl.hadamard(a)*so0 + g1*sd0 + cu*v ) - drte*x[4] - rtete*x[1];
	y[5] = hertegii*( cbcl.hadamard(a)*so0 + g2*sd1 )        - drte*x[5] - rtete*x[2];
	y[8] = hertegii*( cbcl.hadamard(a)*so0 + g3*sd0 )        - drte*x[8] - rtete*x[7];
	y[6] =    hirti*(                        g4*sd7 )        - drti*x[6] - rtiti*x[3];
	for(int i=0;i<DEDIMPERSTATE;i++) { r.paste(y[i],i*sys.states); }
	return r;
}

void Dyneeg::reset(const vector& d)
{
	para = d;	
	int curr = 0;
	
	for(int j=0;j<sys.states;j++) { he(j,DEHEPM*exp(d(curr))); curr++; }
	for(int j=0;j<sys.states;j++) { te(j,DETEPM*exp(d(curr))); curr++; }
	for(int j=0;j<sys.states;j++) { hi(j,DEHIPM*exp(d(curr))); curr++; } 
	for(int j=0;j<sys.states;j++) { ti(j,DETIPM*exp(d(curr))); curr++; }
	dd = DEDDPM*exp(d(curr)); curr++; 
	df = DEDFPM*exp(d(curr)); curr++;
	//t1 = DET1PM*exp(d(curr)); curr++; 
	t2 = DET2PM*exp(d(curr)); curr++;
	for(int j=0;j<sys.dyneeg_flucsorder;j++) { t3(j,d(curr)); curr++; } 
	for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++) { if(j!=k) { cf(j,k,DECFPM*exp(d(curr))); curr++; } } }
	for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++) { if(j!=k) { cb(j,k,DECBPM*exp(d(curr))); curr++; } } }
	for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++) { if(j!=k) { cl(j,k,DECLPM*exp(d(curr))); curr++; } } }
	for(int j=0;j<sys.states;j++) { if(sys.inpc.c_str()[j]=='1')  { cu(j,DECUPM*exp(d(curr))); curr++; } }
	//for(int i=1;i<sys.events;i++) { for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++,curr++) { g[i](j,k,exp(d(curr))); } } }

	setup();
}

void Dyneeg::last(const vector& d)
{
	back.pop_back();
	back.push_front(d);
}

vector Dyneeg::getpara(bool t) const
{
	if(t==0) { return para; }
	else
	{ 
		vector r((3+sys.events)*sys.states*sys.states+5*sys.states+4+sys.dyneeg_flucsorder,ROW);
		int curr = 0;
		for(int j=0;j<sys.states;j++,curr++) { r(curr,he(j)); }
		for(int j=0;j<sys.states;j++,curr++) { r(curr,te(j)); }
		for(int j=0;j<sys.states;j++,curr++) { r(curr,hi(j)); }
		for(int j=0;j<sys.states;j++,curr++) { r(curr,ti(j)); }
		r(curr,dd); curr++; 
		r(curr,df); curr++;
		r(curr,t1); curr++; 
		r(curr,t2); curr++;
		for(int j=0;j<sys.dyneeg_flucsorder;j++,curr++) { r(curr,t3(j)); }
		for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++,curr++) { r(curr,cf(j,k)); } } 
		for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++,curr++) { r(curr,cb(j,k)); } }
		for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++,curr++) { r(curr,cl(j,k)); } }
		for(int j=0;j<sys.states;j++,curr++) { r(curr,cu(j)); }
		for(int i=0;i<sys.events;i++) { for(int j=0;j<sys.states;j++) { for(int k=0;k<sys.states;k++,curr++) { r(curr,g[i](j,k)); } } }
		return r;
	}
}

bool Dyneeg::isstable() const
{
	if((cf-cb-cl).spectrum()(1)<0) { return 1; } else { return 0; }
}
///*

#endif
