///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (dynlin.hh)
///*

///guard
#ifndef HH_DYNLIN
#define HH_DYNLIN
///*

///header
/* class name:	Dynlin
 * 
 * description:	class for linear dynamic system sub model
 * 
 * author:		Christian Himpe
 * 
 * notes:		...
 * 
 * todo:        ...
 * 
 * version:		1.0
 */
///*

///includes
#include "dynbase.hh"
#include "vector.hh"
#include "matrix.hh"
#include "structs.hh"
#include "utils.hh"
///*

///declarations
#define DLDIMPERSTATE 1													//Dynamic Linear Dimensions per State

#define DLOPM   1.0														//Dynamic Linear Sigma Prior Mean
#define DLAPM   0.0														//Dynamic Linear Connectivity Prior Mean
#define DLAPMD -1.0														//Dynamic Linear Connectivity Diagonal Prior Mean
#define DLBPM   0.0														//Dynamic Linear Latent Prior Mean
#define DLCPM   0.0														//Dynamic Linear Input Prior Mean
#define DLCPV   1.0														//Dynamic Linear Input Prior Covariance
///*

///definitions
class Dynlin : public Dynbase
{
	private:
		double o;														//Sigma (scales connectivity and latent matrix)
		matrix a;														//Connectivity Matrix
		matrixarray b;													//Input Induced Connectivity Matrices
		matrix c;														//Input Connectivity Matrix
	
		double ov;														//Sigma Variance
		double abv;														//Connectivity and Input Induced Connectivity Variance
		
		void setup();
	public:
		Dynlin(const syspara& s,void* d);								//Constructor
		~Dynlin() { ; };												//Destructor
		vector initial() { return vector(sys.states,ROW); };			//Initial Values
		vector righthand(const vector& z,const vector& u,double t) const;	//Righthandside
		void   reset(const vector& d);									//Reset Parameters
		void   last(const vector& v) { ; };								//Unused in Linear (MRI) Model
		vector getpara(bool t) const;									//Get Parameters
		bool   isstable() const;										//Check if Linearised System is Stable
};
///*

///implementation
Dynlin::Dynlin(const syspara& s,void* d)
 : Dynbase(s,DLDIMPERSTATE,(1+s.states*s.states+s.states*s.states*s.inputs+s.states*s.inputs)),
   a(s.states,s.states),
   b(s.inputs,matrix(s.states,s.states)),
   c(s.states,s.inputs)
{
	dynlinpara* dl = static_cast<dynlinpara*>(d);
	
	ov *= ov = 1.0/invcumnorm(1.0-0.001);
	//abv = (sys.states*(sys.states-1)) / invcumchi2((sys.states*(sys.states-1)),1.0-0.001);
	//if(sys.states>1) { abv = 0.06/(sys.states-1.8); } else { abv = 0.0; }
	abv = 0.175; //!!!

	if(sys.type==SIM)
	{
		o = dl->o;
		say(o);
		for(int i=0;i<sys.states;i++)
		{
			for(int j=0;j<sys.states;j++)
			{
				a(i,j,dl->a[i*sys.states+j]);
			}
		}

		for(int k=0;k<sys.inputs;k++)
		{
			for(int i=0;i<sys.states;i++)
			{
				for(int j=0;j<sys.states;j++)
				{
					b[k](i,j,dl->b[k][i*sys.states+j]);
				}
			} 
		}

		for(int i=0;i<sys.states;i++)
		{
			for(int j=0;j<sys.inputs;j++)
			{
				c(i,j,dl->c[i*sys.inputs+j]);
			}
		}
	}
	else
	{
		std::vector<double> pm;
		std::vector<double> pc;
		
		o = DLOPM; pm.push_back(o); pc.push_back(ov);
		
		for(int i=0;i<sys.states;i++)
		{
			for(int j=0;j<sys.states;j++)
			{
				if(i==j) { a(i,j,DLAPMD); }
				
				else { a(i,j,DLAPM); pm.push_back(DLAPM); pc.push_back(abv); }
			}
		}
		
		for(int k=0;k<sys.inputs;k++)
		{
			for(int i=0;i<sys.states;i++)
			{
				for(int j=0;j<sys.states;j++)
				{
					b[k](i,j,DLBPM); pm.push_back(DLBPM); pc.push_back(abv);
				}
			}
		}
		
		for(int i=0;i<sys.states;i++)
		{
			for(int j=0;j<sys.inputs;j++)
			{
				c(i,j,DLCPM);
				
				if(i==0 && j==0) { pm.push_back(DLCPM); pc.push_back(DLCPV); }
			}
		}
		
		paras = pm.size();
		vector tp(paras,ROW);
		matrix tc(paras,paras);
		for(int i=0;i<paras;i++) { tp(i,pm[i]); tc(i,i,pc[i]); }
		para = tp;
		cova = tc;
	}

	setup();
}

void Dynlin::setup()
{
	a *= o;
	for(int i=0;i<sys.inputs;i++) { b[i] *= o; }
}

vector Dynlin::righthand(const vector& z,const vector& u,double t) const
{
	vector r(sys.states,ROW);
	r = a*z;
	for(int i=0;i<sys.inputs;i++) { r += b[i]*z*u(i); }
	r += c*u;
	return r;
}

void Dynlin::reset(const vector& d)
{
	para = d;
	int curr = 0;
	
	o = d(curr); curr++;
	
	for(int i=0;i<sys.states;i++)
	{
		for(int j=0;j<sys.states;j++)
		{
			if(i!=j) { a(i,j,d(curr)); curr++; }
		}
	}
	
	for(int k=0;k<sys.inputs;k++)
	{
		for(int i=0;i<sys.states;i++)
		{
			for(int j=0;j<sys.states;j++)
			{
				if(k!=0) { b[k](i,j,d(curr)); curr++; }
			}
		}
	}
	
	for(int i=0;i<sys.states;i++)
	{
		for(int j=0;j<sys.inputs;j++)
		{
			if(i==0 && j==0) { c(i,j,d(curr)); /*curr++;*/ }
		}
	}
	
	setup();
}

vector Dynlin::getpara(bool t) const
{
	if(t==0) { return para; }
	else
	{
		vector r(1+sys.states*sys.states+sys.inputs*sys.states*sys.states+sys.inputs*sys.states,ROW);
		int curr = 0;
			
		r(0,o); curr++;
	
		for(int i=0;i<sys.states;i++)
		{
			for(int j=0;j<sys.states;j++,curr++)
			{
				r(curr,a(i,j)/o);
			}
		}
	
		for(int k=0;k<sys.inputs;k++)
		{
			for(int i=0;i<sys.states;i++)
			{
				for(int j=0;j<sys.states;j++,curr++)
				{
					r(curr,b[k](i,j)/o);
				}
			}
		}
		
		for(int i=0;i<sys.states;i++)
		{
			for(int j=0;j<sys.inputs;j++,curr++)
			{
				r(curr,c(i,j)); 
			}
		}
		
		return r;
	}
}

bool Dynlin::isstable() const
{
	if(a.spectrum()(1)<0) { return 1; } else { return 0; }
}
///*

#endif 
