///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (forhem.hh)
///*

///guard
#ifndef HH_FORHEM
#define HH_FORHEM
///*

///header
/* class name:	Forhem
 * 
 * description:	class for hemodynamic forward system sub model
 * 
 * author:		Christian Himpe
 * 
 * notes:		...
 * 
 * todo:        ...
 * 
 * version:		1.0
 */
///*

///includes
#include <cmath>
#include "vector.hh"
#include "matrix.hh"
#include "structs.hh"
#include "forbase.hh"
///*

///declarations
#define FHDIM 4															//Forward Hemodynamic Dimensions per Output
#define FHPAR 5															//Forward Hemodynamic Parameters per Output

#define FHKPM 0.65														//Forward Hemodynamic Kappa Prior Mean
#define FHKPV 0.015														//Forward Hemodynamic Kappa Prior Covariance
#define FHGPM 0.41														//Forward Hemodynamic Gamma Prior Mean
#define FHGPV 0.002														//Forward Hemodynamic Gamma Prior Covariance
#define FHTPM 0.98														//Forward Hemodynamic Tau Prior Mean
#define FHTPV 0.0568													//Forward Hemodynamic Tau Prior Covariance
#define FHAPM 0.32														//Forward Hemodynamic Alpha Prior Mean
#define FHAPV 0.0015													//Forward Hemodynamic Alpha Prior Covariance
#define FHRPM 0.34														//Forward Hemodynamic Rho Prior Mean
#define FHRPV 0.0024													//Forward Hemodynamic Rho Prior Covariance

#define BVOL 0.02														//Blood Volume Fraction
///*

///definitions
class Forhem : public Forbase
{
	private:
		double kappa;													//Rate of Signal Decay
		double gamma;													//Rate of Flow Dependent Elimination
		double tau;														//Hemodynamic Transit Time
		double alpha;													//Grubbs Exponent
		double rho;														//Resting Oxygen Extraction Fraction
		
		double rtau;													//Reciprocal tau
		double ralpha;													//Reciprocal alpha
		double rrho;													//Reciprocal rho
		
		inline void setup();											//use helper variables for common computations
	public:
		Forhem(const syspara& s,int i,void* f=0);						//Constructor
		~Forhem() { ; };												//Destructor
		vector initial() const;											//Initial Values
		vector righthand(double z,const vector& h);						//Righthandside
		double getoutput(double z,const vector& h) const;				//Output
		void   reset(const vector& f);									//Reset Parameters
};
///*

///implementation
void Forhem::setup()
{
	rtau   = 1.0/tau;
	ralpha = 1.0/alpha;
	rrho   = 1.0/rho;
}

Forhem::Forhem(const syspara& s,int i,void* f) : Forbase(FHDIM,FHPAR,i)
{
	forhempara* fh = static_cast<forhempara*>(f);
	
	if(s.type==SIM)
	{
		kappa = fh->kappa;
		gamma = fh->gamma;
		tau   = fh->tau; 
		alpha = fh->alpha;
		rho   = fh->rho;
	}
	else
	{
		kappa = FHKPM;
		gamma = FHGPM;
		tau   = FHTPM; 
		alpha = FHAPM; 
		rho   = FHRPM;
	}
	
	para(0,kappa); cova(0,0,FHKPV);
	para(1,gamma); cova(1,1,FHGPV);
	para(2,tau);   cova(2,2,FHTPV);
	para(3,alpha); cova(3,3,FHAPV);
	para(4,rho);   cova(4,4,FHRPV);
	
	setup();
}

vector Forhem::initial() const
{
	vector r(dim,ROW);
	r(0,0.0);
	r(1,1.0);
	r(2,1.0);
	r(3,1.0);
	return r;
}

vector Forhem::righthand(double z,const vector& h)
{
	vector r(dim,ROW);
	r(0,z - kappa * h(0) - gamma * ( h(1) - 1.0 ));
	r(1,h(0));
	r(2,rtau * ( h(1) - pow(h(2),ralpha) ));
	r(3,rtau * ( rrho * h(1) * ( 1.0 - pow(1.0-rho,1.0/h(1)) ) - ( pow(h(2),ralpha-1.0) * (h(3)) ) ) );
	return r;
}

double Forhem::getoutput(double z,const vector& h) const
{
	return BVOL * (7.0*rho*(1.0-h(3)) + 2.0*(1.0-h(3)/h(2)) + (2.0*rho - 0.2) * (1.0-h(2)));
}

void Forhem::reset(const vector& f)
{
	kappa = f(0); para(0,kappa);
	gamma = f(1); para(1,gamma);
	tau   = f(2); para(2,tau);
	alpha = f(3); para(3,alpha);
	rho   = f(4); para(4,rho);
	
	setup();
}
///*

#endif 
