///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (import.hh)
///*

///guard
#ifndef HH_IMPORT
#define HH_IMPORT
///*

///header
/* class name:	Import
 * 
 * description:	class for importing datafiles
 * 
 * author:		Christian Himpe
 * 
 * notes:		...
 * 
 * todo:        ...
 * 
 * version:		1.0
 */
///*

///includes
#include <cstdio>
#include <cstring>
#include <string>
#include <map>
#include "structs.hh"
#include "vector.hh"
///*

///declarations
typedef std::pair<vectorarray*,int> datcol;
typedef std::map<int,datcol> junctor;
///*

///definitions
class Import
{
	private:
		syspara sysp;													//General Parameter Structure
		void* dynp;														//Common Pointer to Dynamic Structure 
		void* forp;														//Common Pointer to Forward Structure
		
		dynlinpara* dl;													//Linear MRI Dynamic Structure
		dynnonpara* dn;													//Nonlinear MRI Dynamic Structure
		dyneegpara* de;													//EEG Dynamic Structure
		dyneeepara* df;													//Extended EEG Dynamic Structure
		forhempara* fh;													//Hemodynamic Forward Structure
		foreegpara* fe;													//EEG Forward Structure
		foruntpara* fu;													//Unit Forward Structure
		
		vectorarray input;												//Input Array
		vectorarray data;												//Data Array
		
		void readINI();													//Read INI (Configuration) File
		void readDCM(FILE* f);											//Read DCM (Simulation) File
		void readDAT(FILE* f);											//Read DAT (data Estimation) File
	public:
		Import(const char* fn,const char* ir,int ac);					//Constructor
		~Import() { ; };												//Destructor
		syspara& getsys() { return sysp; }								//Get General Parameter Structure
		vectorarray& getinput() { return input; }						//Get Input Array of Vectors
		vectorarray& getdata() { return data; }							//Get Data Array of Vectors
		void* getdyn() const { return dynp; }							//Get Dynamic System Parameter Structure
		void* getfor() const { return forp; }							//Get Forward System Parameter Structure
		bool gettype() const { return sysp.type; }						//Get Type Calculation
};
///*

///implementation
void Import::readINI()
{
	FILE* f = fopen("dcm.ini","r");
	
	sysp.drift_order = DRIFT_ORDER;
	sysp.feed_datascale = FEED_DATASCALE;
	sysp.dyneeg_intrinsic = DYNEEG_INTRINSIC;
	sysp.dyneeg_flucsorder = DYNEEG_FLUCSORDER;
	sysp.bayes_minloops = BAYES_MINLOOPS;
	sysp.bayes_finitediff = BAYES_FINITEDIFF;
	sysp.bayes_converge = BAYES_CONVERGE;
	sysp.output_sampling = OUTPUT_SAMPLING;
	
	if(f==0) { return; }
	else
	{
		char dummy[64];
		string name;
		for(int i=0;i<11;i++)
		{
			fscanf(f,"%s =",dummy);
			name.assign(dummy);
			if(name.compare("drift_order")==0)       { fscanf(f,"%d",&sysp.drift_order); }
			if(name.compare("feed_datascale")==0)    { fscanf(f,"%lf",&sysp.feed_datascale); }
			if(name.compare("dyneeg_intrinsic")==0)  { fscanf(f,"%lf",&sysp.dyneeg_intrinsic); }
			if(name.compare("dyneeg_flucsorder")==0) { fscanf(f,"%d",&sysp.dyneeg_flucsorder); }
			if(name.compare("bayes_minloops")==0)    { fscanf(f,"%d",&sysp.bayes_minloops); }
			if(name.compare("bayes_finitediff")==0)  { fscanf(f,"%lf",&sysp.bayes_finitediff); }
			if(name.compare("bayes_converge")==0)   { fscanf(f,"%lf",&sysp.bayes_converge); }
			if(name.compare("output_sampling")==0)   { fscanf(f,"%d",&sysp.output_sampling); }
		}
	}
}

void Import::readDCM(FILE* f)
{
	char temp[1024];
	double dummy = 0.0;

	//fscanf(f,"#DCM ID: %*[^\n]%*c",&temp);
	fscanf(f,"#DCM ID: %s \n",temp);
	fscanf(f,"%d %d %d %d %d %lf \n",&sysp.dtype,&sysp.ftype,&sysp.states,&sysp.inputs,&sysp.steps,&sysp.arterr);

	fscanf(f,"#DYNAMIC: \n");
	//Dynamic Parameters
	int s = sysp.states*sysp.states;
	int t = sysp.inputs*sysp.states;
	int u = sysp.inputs;
	int v = sysp.states;
	switch(sysp.dtype)
	{
		case 0: //linear
			dl = new dynlinpara; dynp = static_cast<void*>(dl);
			fscanf(f," %lf \n",&dl->o);
			dl->a = new double[s];
			dl->b = new double*[u]; for(int i=0;i<u;i++) { dl->b[i] = new double[s]; }
			dl->c = new double[t];
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&dl->a[i]); } fscanf(f,"\n");
			for(int i=0;i<u;i++) { for(int j=0;j<s;j++) { fscanf(f,"%lf ",&dl->b[i][j]); } fscanf(f,"\n"); }
			for(int i=0;i<t;i++) { fscanf(f," %lf ",&dl->c[i]); } fscanf(f,"\n");
		break;
		
		case 1: //non-linear
			dn = new dynnonpara; dynp = static_cast<void*>(dn);
			fscanf(f," %lf \n",&dn->o);
			dn->a = new double[s];
			dn->b = new double*[u]; for(int i=0;i<u;i++) { dn->b[i] = new double[s]; }
			dn->c = new double[t];
			dn->d = new double[s];
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&dn->a[i]); } fscanf(f,"\n");
			for(int i=0;i<u;i++) { for(int j=0;j<s;j++) { fscanf(f,"%lf ",&dn->b[i][j]); } fscanf(f,"\n"); }
			for(int i=0;i<t;i++) { fscanf(f," %lf ",&dn->c[i]); } fscanf(f,"\n");
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&dn->d[i]); } fscanf(f,"\n");
		break;

		case 2: //eeg
			de = new dyneegpara; dynp = static_cast<void*>(de);
			de->cf = new double[s];
			de->cb = new double[s];
			de->cl = new double[s];
			de->cu = new double[t];
			de->g  = new double[s];
			fscanf(f," %lf %lf %lf %lf %lf %lf %lf %lf \n",&de->he,&de->hi,&de->te,&de->ti,&de->dd,&de->df,&de->t1,&de->t2);
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&de->cf[i]); } fscanf(f,"\n");
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&de->cb[i]); } fscanf(f,"\n");
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&de->cl[i]); } fscanf(f,"\n");
			for(int i=0;i<t;i++) { fscanf(f," %lf ",&de->cu[i]); } fscanf(f,"\n");
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&de->g[i]); } fscanf(f,"\n");
			fscanf(f," %d",&de->ifo);
			de->ifp = new double[de->ifo];
			for(int i=0;i<de->ifo;i++) { fscanf(f," %lf ",&de->ifp[i]); } fscanf(f,"\n");
		break;
		
		case 3: //eee
			df = new dyneeepara; dynp = static_cast<void*>(df);
			df->cf = new double[s];
			df->cb = new double[s];
			df->cl = new double[s];
			df->cu = new double[t];
			df->g  = new double[s];
			fscanf(f," %lf %lf %lf %lf %lf %lf %lf %lf %lf \n",&df->he,&df->hi,&df->te,&df->ti,&df->ta,&df->dd,&df->df,&df->s1,&df->s2);
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&df->cf[i]); } fscanf(f,"\n");
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&df->cb[i]); } fscanf(f,"\n");
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&df->cl[i]); } fscanf(f,"\n");
			for(int i=0;i<t;i++) { fscanf(f," %lf ",&df->cu[i]); } fscanf(f,"\n");
			for(int i=0;i<s;i++) { fscanf(f," %lf ",&df->g[i]); } fscanf(f,"\n");
		break;
	}
	//*
	
	//Forward Parameters
	fscanf(f,"#FORWARD: \n");
	switch(sysp.ftype)
	{
		case 0: //hemodynamic
			fh = new forhempara; forp = static_cast<void*>(fh);
			fscanf(f," %lf %lf %lf %lf %lf \n",&fh->kappa,&fh->gamma,&fh->tau,&fh->alpha,&fh->rho);
		break;
		
		case 1: //eeg
			fe = new foreegpara; forp = static_cast<void*>(fe);
			fe->lp = new double*[v];
			fe->lm = new double*[v];
			fe->k = new double[v];
			for(int i=0;i<v;i++)
			{
				fe->lp[i] = new double[3];
				fe->lm[i] = new double[3];
				for(int j=0;j<3;j++) { fscanf(f," %lf ",&fe->lp[i][j]); }
				for(int j=0;j<3;j++) { fscanf(f," %lf ",&fe->lm[i][j]); } fscanf(f,"\n");
			}
			for(int i=0;i<v;i++) { fscanf(f," %lf ",&fe->k[i]); } fscanf(f,"\n");
		break;
		
		case 2: //unit scale
			fu = new foruntpara; forp = static_cast<void*>(fu);
			fscanf(f," %lf",&fu->s); fscanf(f,"\n");
		break;
	}
	//*
	
	//Drift
	fscanf(f,"#DRIFT: \n");
	fscanf(f," %d",&sysp.drift_order);
	sysp.dp = new double[(sysp.drift_order+1)*sysp.states];
	for(int i=0;i<(sysp.drift_order+1)*sysp.states;i++) { fscanf(f," %lf ",&sysp.dp[i]); } fscanf(f,"\n");
	//*

	//Inputs	
	fscanf(f,"#INPUTS: \n");
	input.resize(sysp.steps+3,vector(sysp.inputs,ROW));
	int events = 0;
	int start = 0;
	int stop = 0;
	double val = 0.0;
	for(int i=0;i<sysp.inputs;i++)
	{
		fscanf(f," %d \n",&events);
		for(int j=0;j<events;j++)
		{
			fscanf(f," %d>%d=%lf \n",&start,&stop,&val);
			for(int k=start;k<stop;k++) { input[k](i,val); }
		}
	}
	//*

	//One initial data point
	data.push_back(vector(sysp.states,ROW));
	//*
}

void Import::readDAT(FILE* f)
{	
	//determine number of states and inputs (read first line)
	char curr;
	int index;
	int count = 0;
	junctor junction;
	sysp.states = 0;
	sysp.inputs = 0;
	while(true)
	{
		fscanf(f,"\"%c%i\"",&curr,&index);
		switch(curr)
		{
			case 'O': junction[count] = datcol(&data,sysp.states);  sysp.states++; break;
			case 'I': junction[count] = datcol(&input,sysp.inputs); sysp.inputs++; break;
		}
		count++;
		if(fgetc(f)!=',') { break; }
	}
	//*

	//determine step size
	double diff = 0.0;
	fscanf(f,"%lf",&diff);
	fscanf(f,"%*[^\n]%*c");
	fscanf(f,"%lf",&sysp.step);
	sysp.step -= diff;
	//*

	//determine number of steps
	sysp.steps = 1;
	while(fscanf(f,"%*[^\n]%*c")!=EOF) { sysp.steps++; }
	//*

	//read data
	rewind(f);
	fscanf(f,"%*[^\n]%*c");
	data.resize(sysp.steps+1,vector(sysp.states,ROW));
	input.resize(sysp.steps+1,vector(sysp.inputs,ROW));
	int miss[count];
	int last[count];
	double dat = 0.0;
	char dummy;
	fpos_t pos;
	junctor::iterator jit;
	for(int i=0;i<sysp.steps;i++)
	{
		for(int j=0;j<count;j++)
		{
			//read data and check for missing values
			if(j<count-1) { miss[j] = fscanf(f,"%lf",&dat); fgetc(f); }
			else
			{
				fgetpos(f,&pos);
				dummy = fgetc(f);
				if(dummy!='\n')
				{
					fsetpos(f,&pos);		
					 miss[j] = fscanf(f,"%lf",&dat);
				}
				else { miss[j] = 0; }
			} 
			//*
			
			jit = junction.find(j);

			if(jit!=junction.end())
			{
				//interpolate for missing values
				if(last[j]==0 && i>1)
				{ 
					if(miss[j]!=0)
					{
						(*jit->second.first)[i-1](jit->second.second,0.5*((*jit->second.first)[i-2](jit->second.second)+(*jit->second.first)[i](jit->second.second)));
					}
					else
					{
						(*jit->second.first)[i](jit->second.second,0.0);
					}
				}
				if(miss[j]==0 && (i==sysp.steps-1 || i<2)) { dat = 0.0; }
				//*
				
				(*jit->second.first)[i](jit->second.second,dat);

				last[j] = miss[j];
			}
		}
	}
	//*
}

Import::Import(const char* fn,const char* ir,int ac)
 : dynp(0),forp(0),dl(0),dn(0),de(0),fh(0),fe(0),fu(0)
{
	FILE* f = fopen(fn,"r");											//Open Data/Simu File
	if(f==0) { tty(fn); say(" doesn't exist!"); exit(1); }				//Exit if File not Found
	string full(fn);
	int s = full.rfind('/');											//Split Path by last Folder Character
	sysp.name = full.substr(s+1,full.length()-s-5);						//Save Filename
	sysp.path = full.substr(0,s+1);										//Save Path
	readINI();															//Read Confiuration File
	if(full.find(".dcm")!=-1) { sysp.type = SIM; readDCM(f); }
	if(full.find(".mri")!=-1) { sysp.type = EST; sysp.dtype=DYNLIN; sysp.ftype=FORHEM; readDAT(f); }
	if(full.find(".eeg")!=-1) { sysp.type = EST; sysp.dtype=DYNEEG; sysp.ftype=FOREEG; readDAT(f); }
	if(full.find(".eeu")!=-1) { sysp.type = EST; sysp.dtype=DYNEEG; sysp.ftype=FORUNT; readDAT(f); }
	if(full.find(".eee")!=-1) { sysp.type = EST; sysp.dtype=DYNEEE; sysp.ftype=FORUNT; readDAT(f); }
	
	//setup integration stepsizes and sampling length
	if(sysp.dtype==DYNLIN || sysp.dtype==DYNNON) { sysp.istep = MRISTEP; }
	else if(sysp.dtype==DYNEEG || sysp.dtype==DYNEEE) { sysp.istep = EEGSTEP; }
	if(sysp.type==SIM) { sysp.step=sysp.istep; }
	sysp.sample = int(sysp.step/sysp.istep);
	//*

	//setup input configuration
	if(sysp.type==EST)
	{ 
		if(ac>=3) { sysp.inpc = string(ir).substr(0,sysp.states); }
		else { for(int i=0;i<sysp.states;i++) { sysp.inpc +='1'; } }
	}
	//*

	fclose(f);
	tty("[ "); tty(base(fn)); say(" ]");
}
///*

#endif
