///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (matrix.hh)
///*

///guard
#ifndef HH_MATRIX
#define HH_MATRIX
///*

///header
/* class name:	matrix
 * 
 * description:	(sparse) matrix class
 * 
 * author:		Christian Himpe
 * 
 * notes:		...
 * 
 * todo:        ...
 * 
 * version:		1.0
 */
///*

///includes
#include <vector>
#include <map>
#include <tr1/unordered_map>
#include <cmath>
#include <iostream>
#include <cassert>
#include <iomanip>
#include <cstdio>
#include <omp.h>
#include "utils.hh"
///*

///declarations
#define PRACZERO 1.0e-20l 
#define EXPITER 10

class vector;
class matrix;
typedef std::vector<matrix> matrixarray;

#ifdef UN
typedef std::tr1::unordered_map<int,std::tr1::unordered_map<int,double> > sparsematrix;
typedef std::tr1::unordered_map<int,std::tr1::unordered_map<int,double> >::iterator rowiter;
typedef std::tr1::unordered_map<int,std::tr1::unordered_map<int,double> >::const_iterator crowiter;
typedef std::tr1::unordered_map<int,double>::iterator coliter;
typedef std::tr1::unordered_map<int,double>::const_iterator ccoliter;
#else
typedef std::map<int,std::map<int,double> > sparsematrix;
typedef std::map<int,std::map<int,double> >::iterator rowiter;
typedef std::map<int,std::map<int,double> >::const_iterator crowiter;
typedef std::map<int,double>::iterator coliter;
typedef std::map<int,double>::const_iterator ccoliter;
#endif
///*

///definitions
class matrix
{
	private:
		int rows;														//number of rows
		int cols;														//number of coloumns
		sparsematrix sm;												//sparse matrix map	
		matrix cholesky() const;										//cholesky decomposition
	public:
		matrix(int r,int c,double b=0.0,double a=0.0);					//constructor
		~matrix() { ; };												//destructor
		inline double operator()(int r,int c) const __attribute__((fastcall)); //matrix component read
		inline void   operator()(int r,int c,double d);					//matrix component write
		matrix  operator-() const;										//matrix negation
		matrix  operator+(const matrix& m) const;						//matrix additiion
		matrix  operator-(const matrix& m) const;						//matrix subtraction
		matrix  operator*(double x) const; 								//matrix scalar multiplication
		matrix  operator*(const matrix& m) const;						//matrix matrix multiplication
		vector  operator*(const vector& v) const;						//matrix vector multiplication
		matrix& operator+=(const matrix& m);							//matrix assignment addition
		matrix& operator-=(const matrix& m);							//matrix assignment subtraction
		matrix& operator*=(double x);									//matrix assignment scalar multiplication
		matrix& operator=(const matrix& m);								//matrix assignment
		
		int	    dim(bool i) const;										//matrix dimensions
		matrix  copy(int r0,int c0,int rs,int cs) const;				//sub-matrix (from r0,c0 to r0+rs,c0+cs)
		void    paste(const matrix& m,int r,int c);						//copy matrix (m to r,c)
		void    paste(const vector& v,int r,int c);						//copy vector (v to r,c)
		double  trace() const;											//matrix trace
		double  producttrace(const matrix& m) const;					//trace of matrix product
		matrix  transpose() const;										//transpose matrix
		matrix  exponential() const;									//matrixexponential
		matrix  diaginverse() const;									//initial inverse estimation
		matrix  invert() const;											//matrix inversion
		matrix  hadamard(const matrix& m) const;						//matrix hadamrd product
		matrix  kronecker(const matrix& m) const;						//matrix kronecker product
		double  norminf() const;										//matrix infinity norm
		vector  spectrum() const;										//eigenvalue spektrum
		vector  choleskysolver(const vector& v) const;					//solve linear system via cholesky
		matrix  choleskyinverse() const;								//compute inverse via cholesky
		vector  diag() const;											//return diagonal entries of matrix as vector
		void    drop(double x);											//drop entries below given value 
		void    saxpy(const matrix& m,double x);						//saxpy operation (this += x*m) 
		void    clear() { sm.clear(); };								//clear matrix
		void    print() const;											//print matrix
		
		friend matrix operator*(double x,const matrix& m);				//matrix left scalar multiplication
		friend vector operator*(const vector& v,const matrix& m);		//vector matrix multiplication
};
///*

///implementation
matrix matrix::cholesky() const
{
	assert(rows==cols);
	matrix r(rows,cols);
	matrix h(*this);
	double s = 0.0;

	for(int i=0;i<rows-1;i++)
	{
		if(h(i,i)<=0) { say("matrix not positive definite!"); exit(1); }
		r(i,i,sqrt(h(i,i)));
		for(int j=i+1;j<rows;j++)
		{
			r(j,i,h(j,i)/r(i,i));
		}
		for(int j=i+1;j<rows;j++)
		{
			for(int k=j;k<rows;k++)
			{
				h(k,j,h(k,j)-r(k,i)*r(j,i));
			}
		}
	}
	r(rows-1,rows-1,sqrt(h(rows-1,rows-1)));

	return r;
}

matrix::matrix(int r,int c,double b,double a) : rows(r),cols(c),sm()
{
	if(a!=0.0)
	{
		for(int i=0;i<rows;i++) { for(int j=0;j<cols;j++) { if(i!=j) { sm[i][j] = a; } } }
	}
	if(b!=0.0)
	{
		for(int i=0;i<min(rows,cols);++i) { sm[i][i] = b; }
	}
}

double matrix::operator()(int r,int c) const
{
	assert(r<rows && c<cols);
	crowiter i = sm.find(r);
	if(i==sm.end()) { return 0.0; }
	ccoliter j = i->second.find(c);
	if(j==i->second.end()) { return 0.0; }
	return j->second;
}

void matrix::operator()(int r,int c,double d)
{
	assert(r<rows && c<cols);
	if(fabs(d)>PRACZERO) { sm[r][c] = d; return; }
	rowiter i = sm.find(r);
	if(i==sm.end()) { return; }
	i->second.erase(c);
	return; 
}

matrix matrix::operator-() const
{
	matrix r(rows,cols);
	for(crowiter i=sm.begin();i!=sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			r.sm[i->first][j->first] = -(j->second);
		}
	}
	return r;
}

matrix matrix::operator+(const matrix& m) const
{
	assert(rows==m.rows && cols==m.cols);
	matrix r(*this);
	for(crowiter i=m.sm.begin();i!=m.sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			r(i->first,j->first,r(i->first,j->first)+j->second);
		}
	}
	return r;
}

matrix matrix::operator-(const matrix& m) const
{
	assert(rows==m.rows && cols==m.cols);
	matrix r(*this);
	for(crowiter i=m.sm.begin();i!=m.sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			r(i->first,j->first,r(i->first,j->first)-j->second);
		}
	}
	return r;
}

matrix matrix::operator*(double x) const
{
	matrix r(rows,cols);
	//if(fabs(x)<PRACZERO) { return r; }
	for(crowiter i=sm.begin();i!=sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			r(i->first,j->first,j->second*x);
		}
	}
	return r;
}

matrix matrix::operator*(const matrix& m) const
{
	assert(cols==m.rows);
	matrix r(rows,m.cols);
	double t = 0.0;
	crowiter i;
	ccoliter j;
	crowiter a = sm.begin();
	crowiter x = sm.end();
	ccoliter y;
	omp_set_num_threads(threads);
	#pragma omp parallel for private(t,i,j,y) firstprivate(a,x)
	for(int k=0;k<m.cols;++k)
	{
		for(i=a;i!=x;++i)
		{
			t = 0.0;
			y = i->second.end();
			for(j=i->second.begin();j!=y;++j)
			{
				t += j->second * m(j->first,k);
			}
			if(fabs(t)>PRACZERO)
			{
				#pragma omp critical
				{ r.sm[i->first][k] = t; } 
			}
		}
	}
	return r;
}

vector matrix::operator*(const vector& v) const
{
	assert(v.type()==ROW && cols==v.dim());
	vector r(rows,ROW);
	double t = 0.0;
	crowiter i;
	ccoliter j;
	crowiter x = sm.end();
	ccoliter y;
	for(i=sm.begin();i!=x;++i)
	{
		y = i->second.end();
		t = 0.0;
		for(j=i->second.begin();j!=y;++j)
		{
			t += j->second * v(j->first);
		}
		r(i->first,t);
	}
	return r;
}

matrix& matrix::operator+=(const matrix& m)
{
	assert(rows==m.rows && cols==m.cols);
	for(crowiter i=m.sm.begin();i!=m.sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			(*this)(i->first,j->first,(*this)(i->first,j->first)+j->second);
		}
	}
	return *this;
}

matrix& matrix::operator-=(const matrix& m)
{
	assert(rows==m.rows && cols==m.cols);
	for(crowiter i=m.sm.begin();i!=m.sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			(*this)(i->first,j->first,(*this)(i->first,j->first)-j->second);
		}
	}
	return *this;	
}

matrix& matrix::operator*=(double x)
{
	//if(fabs(x)<PRACZERO) { sm.clear(); return *this; }
	for(crowiter i=sm.begin();i!=sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			(*this)(i->first,j->first,j->second * x);
		}
	}
	return *this;
}

matrix& matrix::operator=(const matrix& m)
{
	//assert(rows==m.rows && cols==m.cols);
	sm = m.sm;
	rows = m.rows;
	cols = m.cols;
	return *this;
}

int matrix::dim(bool i) const
{
	if(i==0) { return rows; } else { return cols; }
}

matrix matrix::copy(int r0,int c0,int rs,int cs) const
{
	assert(rows>=rs+r0 && cols>=c0+cs);
	matrix r(rs,cs);
	#ifdef UN
	int a = 0;
	int b = 0;
	for(int i=r0;i<r0+rs;++i,++a)
	{
		b = 0;
		for(int j=c0;j<c0+cs;++j,++b)
		{
			r(a,b,(*this)(i,j));
		}
	}
	#else
	crowiter x = sm.upper_bound(r0+rs-1);
	ccoliter y; 
	for(crowiter i=sm.lower_bound(r0);i!=x;++i)
	{
		y = i->second.upper_bound(c0+cs-1);
		for(ccoliter j=i->second.lower_bound(c0);j!=y;++j)
		{
			r(i->first-r0,j->first-c0,j->second);
		}
	}
	#endif
	return r;
}

void matrix::paste(const matrix& m,int r,int c)
{
	assert(rows>=m.rows+r && cols>=m.cols+c);
	#ifdef UN
	int a = r;
	int b = 0;
	for(int i=0;i<m.rows;++i,++a)
	{
		b = c;
		for(int j=0;j<m.cols;++j,++b)
		{
			(*this)(a,b,m(i,j));
		}
	}
	#else
	rowiter x = sm.upper_bound(r+m.rows-1);
	for(rowiter i=sm.lower_bound(r);i!=x;++i)
	{
		i->second.erase(i->second.lower_bound(c),i->second.upper_bound(c+m.cols-1));
	}
	
	for(crowiter i=m.sm.begin();i!=m.sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			sm[r+i->first][c+j->first] = j->second;
		}
	}
	#endif
}

void matrix::paste(const vector& v,int r,int c)
{
	assert( (v.type()==ROW && rows>=v.dim()+r && cols>=c) ^ (v.type()==COL && cols>=v.dim()+c && rows>=r) );
	if(v.type()==ROW)
	{
		for(int i=0;i<v.dim();++i) { (*this)(r+i,c,v(i)); }
	}
	else
	{
		for(int i=0;i<v.dim();++i) { (*this)(r,c+i,v(i)); }
	}
}

double matrix::trace() const
{
	//assert(rows==cols);
	double r = 0.0;
	for(crowiter i=sm.begin();i!=sm.end();++i)
	{
		r += (*this)(i->first,i->first);
	}
	return r;
}

double matrix::producttrace(const matrix& m) const
{
	assert(cols==m.rows && rows==m.cols);
	double r = 0.0;
	for(crowiter i=sm.begin();i!=sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			r += j->second * m(j->first,i->first);
		}
	}
	return r;
}

matrix matrix::transpose() const
{
	matrix r(cols,rows);
	for(crowiter i=sm.begin();i!=sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			r.sm[j->first][i->first] = j->second;
		}
	}
	return r;
}

matrix matrix::exponential() const
{
	assert(rows==cols);
	matrix e(rows,cols,1.0);
	matrix r(rows,cols,1.0);
	for(int i=1;i<EXPITER;++i)
	{
		e = e*(*this);
		r += (1.0/factorial(i)) * e;
	}
	return r;
}

matrix matrix::diaginverse() const
{
	assert(rows==cols);
	matrix r(cols,rows);
	for(int i=0;i<rows;i++)
	{
		r(i,i,1.0/(*this)(i,i));
	}
	return r;
}

matrix matrix::invert() const
{
	//assert(rows==cols);
	
	//initial 
	matrix r(transpose());
	vector s((*this*r).spectrum());
	double o = 1.0/(norminf() * r.norminf());
	r *= o;
	int t = int(ceil(2.0*log(fabs(s(1)/s(0)))))+1;
	say("");
	tty(t);
	//*
	
	//Newton (Hotelling-Bodewig)
	matrix h(rows,cols,2.0);
	for(int i=0;i<t;++i) { r = r*(h - (*this) * r); tty("~"); }
	//*
	return r;
}

matrix matrix::hadamard(const matrix& m) const
{
	assert(rows==m.rows && cols==m.cols);
	matrix r(rows,cols);
	for(crowiter i=sm.begin();i!=sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			r(i->first,j->first,j->second * m(i->first,j->first));
		}
	}
	return r;
}

matrix matrix::kronecker(const matrix& m) const
{
	matrix r(rows*m.rows,cols*m.cols);
	for(crowiter i=sm.begin();i!=sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			r.paste(j->second*m,i->first*m.rows,j->first*m.cols);
		}
	}
	return r;
}

double matrix::norminf() const
{
	double r = 0.0;
	double t = 0.0;
	for(crowiter i=sm.begin();i!=sm.end();++i)
	{
		t = 0.0;
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			t += fabs(j->second);
		}
		r = max(r,t);
	}
	return r;
}

vector matrix::spectrum() const
{
	assert(rows==cols);
	vector r(2,ROW);
	double t = 0.0;
	double u = 0.0;
	double v = 0.0;
	bool f = 1;
	for(crowiter i=sm.begin();i!=sm.end();++i)
	{
		t = 0.0;
		u = v = (*this)(i->first,i->first);
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			t += fabs(j->second);
		}
		t -= fabs(u);
		u -= t;
		v += t;
		if(f==1)
		{
			f = 0;
			r(0,u);
			r(1,v);
		}
		r(0,min(r(0),u));
		r(1,max(r(1),v));
	}
	return r;
}

vector matrix::choleskysolver(const vector& v) const
{
	assert(rows==cols);
	matrix l(cholesky());
	vector r(rows,ROW);
	double s = 0.0;

	for(crowiter i=l.sm.begin();i!=l.sm.end();++i)
	{
		s = v(i->first);
		for(ccoliter k=i->second.begin();k!=i->second.end();++k)
		{
			s -= k->second * r(k->first);
		}
		r(i->first,s/l(i->first,i->first));
	}

	for(int i=rows-1;i>=0;--i)
	{
		s = r(i);
		for(int k=i+1;k<rows;++k)
		{
			s -= l(k,i) * r(k);
		}
		r(i,s/l(i,i));
	}
	return r;
}

matrix matrix::choleskyinverse() const
{
	assert(rows==cols);
	matrix l(cholesky());
	matrix r(rows,cols,1.0);
	double s = 0.0;

	//omp_set_num_threads(threads);
	//#pragma omp parallel for firstprivate(s)
	for(int i=0;i<cols;i++)
	{
		for(int j=0;j<rows;j++)
		{
			s = r(j,i);
			for(int k=0;k<j;k++)
			{
				s -= l(j,k) * r(k,i);
			}
			//#pragma omp critical
			{ r(j,i,s/l(j,j)); }
		}
	}
	r = r.transpose()*r;
	return r;
}

vector matrix::diag() const
{
	assert(rows==cols);
	vector r(rows,ROW);
	for(crowiter i=sm.begin();i!=sm.end();++i) { r(i->first,(*this)(i->first,i->first)); }
	return r;
}

void matrix::saxpy(const matrix& m,double x)
{
	assert(rows==m.rows && cols==m.cols);
	for(crowiter i=m.sm.begin();i!=m.sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			(*this)(i->first,j->first,(*this)(i->first,j->first)+x*j->second);
		}
	}	
}

void matrix::drop(double x)
{
	for(rowiter i=sm.begin();i!=sm.end();++i)
	{
		for(coliter j=i->second.begin();j!=i->second.end();++j)
		{
			if(fabs(j->second)<x) { i->second.erase(j->first); }
		}
	}	
}

void matrix::print() const
{
	std::cout.precision(3);
	if(rows!=1) { std::cout << "/ "; } else { std::cout << "< "; }
	for(int j=0;j<cols;j++)
	{
		if(sm.find(0)==sm.end() || sm.find(0)->second.find(j)==sm.find(0)->second.end()) { std::cout << std::setfill(' ') << std::fixed << std::setw(6) << 0 << " "; }
		else { std::cout << std::setfill(' ') << std::fixed << std::setw(6) << sm.find(0)->second.find(j)->second << " "; }
	}

	if(rows!=1) { std::cout << " \\" << std::endl; } else { std::cout << " >"; } 
	for(int i=1;i<rows-1;i++)
	{
		std::cout << "| ";
		for(int j=0;j<cols;j++)
		{
			if(sm.find(i)==sm.end() || sm.find(i)->second.find(j)==sm.find(i)->second.end()) { std::cout << std::setfill(' ') << std::fixed << std::setw(6) << 0 << " "; }
			else { std::cout << std::setfill(' ') << std::fixed << std::setw(6) << sm.find(i)->second.find(j)->second << " "; }
		}
		std::cout << " |" << std::endl; 
	}

	if(rows!=1)
	{
		std::cout << "\\ ";
		for(int j=0;j<cols;j++)
		{
			if(sm.find(rows-1)==sm.end() || sm.find(rows-1)->second.find(j)==sm.find(rows-1)->second.end()) { std::cout << std::setfill(' ') << std::fixed << std::setw(6) << 0 << " "; }
			else { std::cout << std::setfill(' ') << std::fixed << std::setw(6) << sm.find(rows-1)->second.find(j)->second << " "; }
		}
		std::cout << " /" << std::endl;
	}
	std::cout << std::endl; 
}

matrix operator*(double x,const matrix& m)
{
	matrix r(m.rows,m.cols);
	for(crowiter i=m.sm.begin();i!=m.sm.end();++i)
	{
		for(ccoliter j=i->second.begin();j!=i->second.end();++j)
		{
			r(i->first,j->first,j->second * x);
		}
	}
	return r;
}

vector operator*(const vector& v,const matrix& m)
{
	assert(v.rc==COL && v.size==m.rows);
	vector r(m.cols,COL);
	double t = 0.0;
	//omp_set_num_threads(threads);
	//#pragma omp parallel for private(t)
	for(int j=0;j<m.cols;j++)
	{
		t = 0.0;
		for(int i=0;i!=v.size;++i)
		{
			t += v(i)*m(i,j);
		}
		//#pragma omp critical
		{ r(j,t); }
	}
	return r;
}

matrix dyadic(const vector& v,const vector& w)
{
	assert(v.size==w.size);
	matrix r(v.size,w.size);
	for(int i=0;i<v.size;i++)
	{
		for(int j=0;j<w.size;++j)
		{
			r(i,j,v(i) * w(j));
		}
	}
	return r;
}

matrix diag(const vector& v)
{
	matrix r(v.size,v.size);
	for(int i=0;i<v.size;i++)
	{
		r(i,i,v(i));
	}
	return r;
}
///*

#endif 
