///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (output.hh)
///*

///guard
#ifndef HH_OUTPUT
#define HH_OUTPUT
///*

///header
/* class name:	Output
 * 
 * description:	class for managing and writing output
 * 
 * author:		Christian Himpe
 * 
 * notes:		...
 * 
 * todo:        ...
 * 
 * version:		1.0
 */
///*

///includes
#include <cstdio>
#include <string>
#include <cstdlib>
#include <vector>
#include "vector.hh"
#include "structs.hh"
#include "system.hh"
#include "feed.hh"
#include "drift.hh"
#include "utils.hh"
///*

///definitions
class Output
{
	private:
		const syspara& sys;												//General Parameter Structure
		System& system;													//System Superclass
		Feed& feed;														//Data and Input Provider
		Drift& drift;													//Drift Provider
		vectorarray ddata;
		vectorarray odata;
		vector err;
		vector coi;
		string datafile;												//Filename of Data
		string plotfile;												//Filename of Gnuplot 
		string flowfile;												//Filename of Flow
		string parafile;												//Filename of Parameter List
	public:
		Output(const syspara& s,System& x,Feed& f,Drift& d);			//Constructor
		~Output() { ; };												//Destructor
		void set(int i,const vector& z,const vector& o);				//Set Data
		void error();													//Compute Error
		void writeDAT(); //const;											//Write Data File
		void writePLT() const;											//Write Plot File
		void writeFLW() const;											//Write Flow File
		void writeTXT() const;											//Write Para File
};
///*

///implementation
Output::Output(const syspara& s,System& x,Feed& f,Drift& d)
 : sys(s),
   system(x),
   feed(f),
   drift(d),
   ddata(s.steps,vector(s.states,ROW)),
   odata(s.steps,vector(s.states,ROW)),
   err(s.steps,ROW),
   coi(s.steps,ROW)
{
	string ext;
	if(sys.ftype==FORHEM) { ext=".mri"; }
	if(sys.ftype==FOREEG) { ext=".eeg"; }
	if(sys.dtype==DYNEEG && sys.ftype==FORUNT) { ext=".eeu"; } 
	if(sys.dtype==DYNEEE && sys.ftype==FORUNT) { ext=".eee"; } 
	datafile = (sys.path+sys.name+"/"+sys.name+ext);
	plotfile = (sys.path+sys.name+"/"+sys.name+".plt");
	flowfile = (sys.path+sys.name+"/"+sys.name+".flw");
	parafile = (sys.path+sys.name+"/"+sys.name+".txt");
	string dir = "mkdir -p "+sys.path+sys.name;
	::system(dir.c_str());
	ok();
}

void Output::set(int i,const vector& z,const vector& o)
{
	ddata[i] = z;
	odata[i] = o;
}

void Output::error()
{
	double eout = 0.0;
	if(sys.type==EST)
	{
		double omax = feed.data(0)(0);
		double omin = feed.data(0)(0);
		for(int i=0;i<sys.steps;i++)
		{
			err(i,(feed.data(i)-odata[i]).length());
			eout += err(i);
			for(int j=0;j<sys.states;j++)
			{
				omax = max(omax,feed.data(i)(j));
				omin = min(omin,feed.data(i)(j));
			}
		}
		eout *= (1.0/sys.steps) * (100.0/(omax-omin));
	}
	else
	{
		eout = pow(sys.istep,5);
		for(int i=0;i<sys.steps;i++)
		{
			err(i,eout);
		}
	}
	std::cout.precision(1);
	std::cout << "[   " << std::setw(4) << std::setfill(' ') << eout << "%  ]" << std::endl;
}

void Output::writeDAT() //const
{
	double v = 1.0;
	if(sys.type==SIM) { v = 1000.0; }
	
	FILE* f = fopen(datafile.c_str(),"w");

	//header
	fprintf(f,"\"T0\",");
	fprintf(f,"\"E0\","); 
	for(int j=0;j<sys.inputs;j++) { fprintf(f,"\"I%i\",",j); } 
	if(sys.dtype==DYNEEG) { for(int j=0;j<sys.inputs;j++) { fprintf(f,"\"J%i\",",j); } }
	for(int j=0;j<sys.states;j++) { fprintf(f,"\"Z%i\",",j); } 
	for(int j=0;j<sys.states;j++) { fprintf(f,"\"O%i\"",j); if(j!=sys.states-1 || sys.type==EST) { fprintf(f,","); } } 
	if(sys.type==EST) { for(int j=0;j<sys.states;j++) { fprintf(f,"\"M%i\"",j); if(j!=sys.states-1) { fprintf(f,","); } } } 
	fprintf(f,"\n");
	//*
	
	//Generate EEG input
	if(sys.dtype==DYNEEG || sys.dtype==DYNEEE)
	{
		vector p(system.getpara(1));
		int q = 4*sys.states+2;
		vector g(sys.dyneeg_flucsorder+2,ROW);
		g(0,p(q));
		g(1,p(q+1));
		for(int i=0;i<sys.dyneeg_flucsorder;i++) { g(2+i,p(q+2+i)); }
		
		double r;
		double t;
		double e;
		for(int i=0;i<sys.steps;i++)
		{
			e = double(i*sys.sample)*sys.istep; 
			t = (e - feed.jnput(e)(1));
			r = pow(g(1),g(0))/gamma(g(0))*pow(t,g(0)-1.0)*exp(-g(1)*t);
			for(int j=0;j<sys.dyneeg_flucsorder;j++) { r += g(2+j)*cos(2.0*M_PI*double(j)*t); }
			coi(i,sign(feed.jnput(e)(0))*r);
		}
	}
	//*
	
	//body
	vector iout(sys.inputs,ROW);
	vector nout(sys.states,ROW);
	vector oout(sys.states,ROW);
	int ostep = 1; if(sys.type==SIM) { ostep = sys.output_sampling; }
	
	for(int i=0;i<sys.steps;i+=ostep)	 
	{
		for(int j=0;j<ostep;j++)
		{
			iout += feed.input(i+j);
			nout += ddata[i+j];
			oout += odata[i+j];
		}
		iout *= 1.0/double(ostep);
		nout *= 1.0/double(ostep);
		oout *= 1.0/double(ostep);
		
		fprintf(f,"%f,",i*sys.step);
		fprintf(f,"%f,",err(i));
		for(int j=0;j<sys.inputs;j++) { fprintf(f,"%f,",iout(j)); }
		if(sys.dtype==DYNEEG) { for(int j=0;j<sys.inputs;j++) { fprintf(f,"%f,",coi(i)); } }
		for(int j=0;j<sys.states;j++) { fprintf(f,"%f,",nout(j)); }
		for(int j=0;j<sys.states;j++) { fprintf(f,"%f",v*oout(j)); if(j!=sys.states-1 || sys.type==1) { fprintf(f,","); } }
		if(sys.type==EST) { for(int j=0;j<sys.states;j++) { fprintf(f,"%f",feed.data(i)(j)); if(j!=sys.states-1) { fprintf(f,","); } } }
		fprintf(f,"\n");
		
		iout.clear();
		nout.clear();
		oout.clear();
	}
	//*
	
	fclose(f);
	ok();
}

void Output::writePLT() const
{
	FILE* f = fopen(plotfile.c_str(),"w");
	
	int curr = 3;
	
	//header
	fprintf(f,"#!/usr/bin/gnuplot -persist \n");
	fprintf(f,"unset mouse \n");
	fprintf(f,"set term x11 \n");
	fprintf(f,"set datafile separator \",\" \n");
	fprintf(f,"set multiplot layout %i,1 title \"DCM Output ( %s ) - Simu\" \n",sys.inputs+sys.states+sys.states+2,sys.name.c_str());
	//*
	
	//plot config
	fprintf(f,"set xlabel \"\" \n");
	fprintf(f,"set xrange [%f:%f] \n",0.0,double(sys.steps*sys.step));
	fprintf(f,"set nokey \n");
	fprintf(f,"set bmargin 0 \n");
	fprintf(f,"set tmargin 0.1 \n");
	fprintf(f,"set lmargin 12 \n");
	fprintf(f,"set border 10 \n");
	fprintf(f,"set ytics \n");
	fprintf(f,"set xtics nomirror \n");
	fprintf(f,"set noxtics \n");
	fprintf(f,"set format y \"%%2.6f\" \n");
	//*
	
	//draw inputs
	fprintf(f,"\n");
	fprintf(f,"# Inputs: \n");
	for(int i=0;i<sys.inputs;i++)
	{
		fprintf(f,"set ylabel \"Input[%i]\" \n",i);
		fprintf(f,"plot \"%s\" index 0 using 1:%i with steps",datafile.c_str(),i+curr);
		if(sys.dtype==DYNEEG || sys.dtype==DYNEEE)
		{
			fprintf(f,", \"%s\" index 0 using 1:%i with lines \n",datafile.c_str(),i+curr+sys.inputs);
		}
		fprintf(f,"\n");
	}
	if(sys.dtype==DYNEEG) { curr+=sys.inputs; }		
	//*
	
	//draw neuronal
	curr += sys.inputs;
	fprintf(f,"\n");
	fprintf(f,"# Neuronals: \n");
	for(int i=0;i<sys.states;i++)
	{
		fprintf(f,"set ylabel \"Neuronal[%i]\" \n",i);
		fprintf(f,"plot \"%s\" index 0 using 1:%i with lines \n",datafile.c_str(),i+curr);
	}	
	//*
	
	//draw forward
	curr += sys.states;
	fprintf(f,"\n");
	fprintf(f,"# Forwards: \n");
	for(int i=0;i<sys.states;i++)
	{
		fprintf(f,"set ylabel \"Forward[%i]\" \n",i);
		fprintf(f,"plot \"%s\" index 0 using 1:%i with lines",datafile.c_str(),i+curr);
		if(sys.type==EST) {fprintf(f,", \"%s\" index 0 using 1:%i with lines",datafile.c_str(),i+sys.states+curr); }
		fprintf(f,"\n");
	}
	//*
	
	//draw error
	curr += sys.states;
	fprintf(f,"\n");
	fprintf(f,"# Error: \n");
	fprintf(f,"set ylabel \"Error\" \n");
	fprintf(f,"plot \"%s\" index 0 using 1:2 with lines \n",datafile.c_str());		
	//*
	
	//draw empty graph for xaxis
	fprintf(f,"set size 1,.1 \n");
	fprintf(f,"set notitle \n");
	fprintf(f,"set ylabel \" \" \n");
	fprintf(f,"set yrange [0:1] \n");
	fprintf(f,"set xtics nomirror \n");
	fprintf(f,"set tics out \n");
	fprintf(f,"set noytics \n");
	fprintf(f,"set bmargin 2.0 \n");
	fprintf(f,"set border 1 \n");
	fprintf(f,"plot -1 \n");
	//*
	
	//footer
	fprintf(f,"\n");
	fprintf(f,"set nomultiplot \n");
	fprintf(f,"#pause -1 \n");
	fprintf(f," \n");
	//*

	fclose(f);
	ok();
}

void Output::writeFLW() const
{
	FILE* f = fopen(flowfile.c_str(),"w");
	
	//header
	fprintf(f,"digraph %s { \n",sys.name.c_str());
	//*
	
	//setup regions
	fprintf(f,"node [shape=circle];");
	for(int i=0;i<sys.states;i++)
	{
	    fprintf(f,"z%i",i);
	    if(i<sys.states-1) { fprintf(f,";"); }
	}
	fprintf(f," \n");
	//*
	
	//setup inputs
	fprintf(f,"node [shape=box];");
	for(int i=0;i<sys.inputs;i++)
	{
	    fprintf(f,"u%i",i);
	    if(i<sys.inputs-1) { fprintf(f,";"); }
	}
	fprintf(f,"\n");
	//*
	
	//~ int aff = 0;
	//~ dynlinpara* dl;
	//~ dynnonpara* dn;
	//~ dyneegpara* de;
	//~ switch(dtype)
	//~ {
		//~ //linear
		//~ case 0:
			//~ dl = static_cast<dynlinpara*>(d);
			//~ //connections regions to regions
			//~ for(int i=0;i<states;i++)
			//~ {
				//~ for(int j=0;j<states;j++)
				//~ {
					//~ if(fabs(dl->a[(i*states)+j])>0.1e-10 && i!=j)
					//~ {
						//~ aff = 0;
						//~ for(int k=0;k<inputs;k++) { aff+=(fabs(dl->b[k][(i*states)+j])>0.1e-10); }
						//~ if(aff!=0)
						//~ {
							//~ fprintf(f,"c%i%i [shape=none,label=\"%.2f\",height=.01,width=.01]; \n",j,i,dl->a[(i*states)+j]);
							//~ fprintf(f,"z%i->c%i%i [dir=none,weight=1]; \n",j,j,i);
							//~ fprintf(f,"c%i%i->z%i [weight=1]; \n",j,i,i);
						//~ }
						//~ else
						//~ {
							//~ fprintf(f,"z%i->z%i [weight=3,label=\"%.2f\"]; \n",j,i,dl->a[(i*states)+j]);
						//~ }
					//~ }
				//~ }
			//~ }
			//~ //*
			//~ 
			//~ //connections inputs to regions
			//~ for(int i=0;i<states;i++)
			//~ {
				//~ for(int j=0;j<inputs;j++)
				//~ {
					//~ if(fabs(dl->c[(i*inputs)+j])>0.1e-10)
					//~ {
						//~ fprintf(f,"u%i->z%i [weight=1,label=\"%.2f\"]; \n",j,i,dl->c[(i*inputs)+j]);
					//~ }
				//~ }
			//~ }
			//~ //*
			//~ 
			//~ //connections inputs to connections
			//~ for(int k=0;k<inputs;k++)
			//~ {
				//~ for(int i=0;i<states;i++)
				//~ {
					//~ for(int j=0;j<states;j++)
					//~ {
						//~ if(fabs(dl->b[k][(i*states)+j])>0.1e-10)
						//~ {
							//~ if(i!=j)
							//~ {
								//~ fprintf(f,"u%i->c%i%i [weight=1,label=\"%.2f\"]; \n",k,j,i,dl->b[k][(i*states)+j]);
							//~ }
							//~ else
							//~ {
								//~ fprintf(f,"u%i->z%i [label=\"%.2f\"]; \n",k,i,dl->b[k][(i*states)+j]);
							//~ }
						//~ }
					//~ }
				//~ }
			//~ }
			//~ //*
		//~ break;
		//~ //*
//~ 
		//~ //non-linear
		//~ case 1:
			//~ dn = static_cast<dynnonpara*>(d);
			//~ //connections regions to regions
			//~ for(int i=0;i<states;i++)
			//~ {
				//~ for(int j=0;j<states;j++)
				//~ {
					//~ if(fabs(dn->a[(i*states)+j])>0.1e-10 && i!=j)
					//~ {
						//~ aff = 0;
						//~ for(int k=0;k<inputs;k++) { aff+=(fabs(dn->b[k][(i*states)+j])>0.1e-10); }
						//~ if(aff!=0)
						//~ {
							//~ fprintf(f,"c%i%i [shape=diamond,style=filled,label=\"\",height=.1,width=.1]; \n",j,i);
							//~ fprintf(f,"z%i->c%i%i [dir=none,weight=1]; \n",j,j,i);
							//~ fprintf(f,"c%i%i->z%i [weight=1]; \n",j,i,i);
						//~ }
						//~ else
						//~ {
							//~ fprintf(f,"z%i->z%i [weight=3]; \n",j,i);
						//~ }
					//~ }
				//~ }
			//~ }
			//~ //*
			//~ 
			//~ //connections inputs to regions
			//~ for(int i=0;i<states;i++)
			//~ {
				//~ for(int j=0;j<inputs;j++)
				//~ {
					//~ if(fabs(dn->c[(i*inputs)+j])>0.1e-10)
					//~ {
						//~ fprintf(f,"u%i->z%i [weight=1]; \n",j,i);
					//~ }
				//~ }
			//~ }
			//~ //*
			//~ 
			//~ //connections inputs to connections
			//~ for(int k=0;k<inputs;k++)
			//~ {
				//~ for(int i=0;i<states;i++)
				//~ {
					//~ for(int j=0;j<states;j++)
					//~ {
						//~ if(fabs(dn->b[k][(i*states)+j])>0.1e-10)
						//~ {
							//~ if(i!=j)
							//~ {
								//~ fprintf(f,"u%i->c%i%i [weight=1]; \n",k,j,i);
							//~ }
							//~ else
							//~ {
								//~ fprintf(f,"u%i->z%i; \n",k,i);
							//~ }
						//~ }
					//~ }
				//~ }
			//~ }
			//~ //*
			//~ 
			//~ //connections states to connections
			//~ //...
			//~ //*
		//~ break;
		//~ //*
		//~ 
		//~ //eeg
		//~ case 2:
			//~ de = static_cast<dyneegpara*>(d);
			//~ //...
		//~ break;
		//~ //*
	//~ }
	//~ 
	//~ //footer
	//~ fprintf(f,"overlap=false \n");
	//~ //fprintf(file,"label=\"%s\" \n",name);
	//~ fprintf(f,"} \n");
	//~ //*
	
	fclose(f);
	ok();
}

void Output::writeTXT() const
{
	FILE* f = fopen(parafile.c_str(),"w");
	
	vector p(system.getpara(1));

	int curr = 0;
	
	//Dynamic Parameters
	switch(sys.dtype)
	{
		case DYNLIN:
			fprintf(f,"Dynamic Parameters (%s): \n","Hemodynamic Linear");
			fprintf(f,"o: %f \n",p(curr)); curr++;
			for(int i=0;i<sys.states;i++)
			{
				for(int j=0;j<sys.states;j++,curr++)
				{
					fprintf(f,"a%i%i: %f \n",i,j,p(curr));
				}
			}
			for(int l=0;l<sys.inputs;l++)
			{
				for(int i=0;i<sys.states;i++)
				{
					for(int j=0;j<sys.states;j++,curr++)
					{
						fprintf(f,"b[%i]%i%i: %f \n",l,i,j,p(curr));
					}
				}
			}
			for(int i=0;i<sys.states;i++)
			{
				for(int j=0;j<sys.inputs;j++,curr++)
				{
					fprintf(f,"c%i%i: %f \n",i,j,p(curr));
				}
			}
		break;
		
		case DYNNON:
			fprintf(f,"Dynamic Parameters (%s): \n","Hemodynamic Nonlinear");
			
		break;
		
		case DYNEEG:
			fprintf(f,"Dynamic Parameters (%s): \n","Eeg I");
			for(int i=0;i<sys.states;i++,curr++) { fprintf(f,"He[%i]: %f \n",i,p(curr)); }
			for(int i=0;i<sys.states;i++,curr++) { fprintf(f,"Te[%i]: %f \n",i,p(curr)); }
			for(int i=0;i<sys.states;i++,curr++) { fprintf(f,"Hi[%i]: %f \n",i,p(curr)); }
			for(int i=0;i<sys.states;i++,curr++) { fprintf(f,"Ti[%i]: %f \n",i,p(curr)); }
			fprintf(f,"D0: %f \n",p(curr)); curr++;
			fprintf(f,"D1: %f \n",p(curr)); curr++;
			fprintf(f,"t1: %f \n",p(curr)); curr++;
			fprintf(f,"t2: %f \n",p(curr)); curr++;
			for(int i=0;i<sys.dyneeg_flucsorder;i++,curr++) { fprintf(f,"t3[%i]: %f \n",i,p(curr)); }
			for(int i=0;i<sys.states;i++)
			{
				for(int j=0;j<sys.states;j++,curr++) { fprintf(f,"Cf[%i,%i]: %f \n",i,j,p(curr)); }
			}
			for(int i=0;i<sys.states;i++)
			{
				for(int j=0;j<sys.states;j++,curr++) { fprintf(f,"Cb[%i,%i]: %f \n",i,j,p(curr)); }
			}
			for(int i=0;i<sys.states;i++)
			{
				for(int j=0;j<sys.states;j++,curr++) { fprintf(f,"Cl[%i,%i]: %f \n",i,j,p(curr)); }
			}
			for(int i=0;i<sys.states;i++,curr++) { fprintf(f,"Cu[%i]: %f \n",i,p(curr)); }
			for(int k=0;k<sys.events;k++)
			{
				for(int i=0;i<sys.states;i++)
				{
					for(int j=0;j<sys.states;j++,curr++) { fprintf(f,"G[%i,%i,%i]: %f \n",k,i,j,p(curr)); }
				}
			}
		break;
		
		case DYNEEE:
			fprintf(f,"Dynamic Parameters (%s): \n","Eeg II");
		break;
	}
	//*

	//Forward Parameters
	switch(sys.ftype)
	{
		case 0:
			fprintf(f,"Forward Parameters (%s): \n","Hemodynamic");
			fprintf(f,"K: %f \n",p(curr)); curr++;
			fprintf(f,"G: %f \n",p(curr)); curr++;
			fprintf(f,"T: %f \n",p(curr)); curr++;
			fprintf(f,"A: %f \n",p(curr)); curr++;
			fprintf(f,"R: %f \n",p(curr)); curr++;
		break;
		
		case 1:
			fprintf(f,"Forward Parameters (%s): \n","Eeg");
			for(int i=0;i<sys.states;i++,curr++)
			{
				fprintf(f,"K[%i]: %f \n",i,p(curr));
				fprintf(f,"Lp[%i]x: %f \n",i,p(curr));
				fprintf(f,"Lp[%i]y: %f \n",i,p(curr));
				fprintf(f,"Lp[%i]z: %f \n",i,p(curr));
				fprintf(f,"Lm[%i]x: %f \n",i,p(curr));
				fprintf(f,"Lm[%i]y: %f \n",i,p(curr));
				fprintf(f,"Lm[%i]z: %f \n",i,p(curr));
			}
		break;
		
		case 2:
			fprintf(f,"Forward Parameters (%s): \n","Unit Scale");
			for(int i=0;i<sys.states;i++,curr++)
			{
				fprintf(f,"S[%i]: %f \n",i,p(curr));
			}
	}
	//*
	
	//Drift Parameters
	if(sys.drift_order!=-1)
	{
		vector dp(drift.getpara());
		int ord = dp.dim()/sys.states;
		fprintf(f,"Drift Parameters: (Order:%i)\n",ord-1);
		for(int i=0;i<ord;i++)
		{
			for(int j=0;j<sys.states;j++)
			{
				fprintf(f,"OS[%i,%i]: %f \n",i,j,dp(sys.states*i+j));
			}
		}
	}
	//*

	fclose(f);
	ok();
}
///*

#endif
