///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (solver.hh)
///*
 
///guard
#ifndef HH_SOLVER
#define HH_SOLVER
///*

///header
/* class name:	Solver
 * 
 * description:	differential equation solver class using cvode
 * 
 * author:		Christian Himpe
 * 
 * notes:		...
 * 
 * todo:        ...
 * 
 * version:		1.0
 */
///*

///includes
#include <cmath>
#include "structs.hh"
#include "vector.hh"
#include "system.hh"
#include "output.hh"
#include "drift.hh"
#include "utils.hh"
///*

///declarations

///*

///definitions
class Solver
{
	private:
		const syspara& sys;												//General Parameter Structure
		System& system;													//Reference to System Class
		Drift& drift;													//Drift Provider
		
		int steps;														//Number of Full Steps 
		int dim;														//Dimension of Complete System
	public:
		Solver(const syspara& s,System& x,Drift& d,bool b=0);			//Constructor
		~Solver() { ; };												//Destructor
		vector solve() const;											//Solver (Bayes)
		void solve(Output& o);											//Solver (Output)
};
///*

///implementation

Solver::Solver(const syspara& s,System& x,Drift& d,bool b)
 : sys(s),
   system(x),
   drift(d),
   dim(x.getdim())
{
	steps = int(sys.steps*sys.sample);
	if(b!=0) { tty("[    "); tty("5"); say(".    ]"); }
}

vector Solver::solve() const
{
	vector y(system.initial());
	vector r(sys.states*sys.steps,ROW);
	vector k0(y.dim(),ROW);
	vector k1(y.dim(),ROW);
	vector k2(y.dim(),ROW);
	vector k3(y.dim(),ROW);
	vector k4(y.dim(),ROW);
	vector k5(y.dim(),ROW);
	vector a(sys.states,ROW);
	double t = 0.0;

	for(int i=0;i<steps;i++,t+=sys.istep)
	{
		k0 = sys.istep * system.righthand(y,t);
		k1 = sys.istep * system.righthand(y + (1.0/4.0) * k0,t + (1.0/4.0) * sys.istep);
		k2 = sys.istep * system.righthand(y + (3.0/32.0) * k0 + (9.0/32.0) * k1,t + (3.0/8.0) * sys.istep);
		k3 = sys.istep * system.righthand(y + (1932.0/2197.0) * k0 - (7200.0/2197.0) * k1 + (7296.0/2197.0) * k2,t + (12.0/13.0) * sys.istep);
		k4 = sys.istep * system.righthand(y + (439.0/216.0) * k0 - 8.0 * k1 + (3680.0/513.0) * k2 - (845.0/4104.0) * k3,t + sys.istep);
		k5 = sys.istep * system.righthand(y - (8.0/27.0) * k0 + 2.0 * k1 - (3544.0/2565.0) * k2 + (1859.0/4104.0) * k3 - (11.0/40.0) * k4,t + (1.0/2.0) * sys.istep);
		y += (16.0/135.0)*k0 + (6656.0/12825.0)*k2 + (28561.0/56430.0)*k3 - (9.0/50.0)*k4 + (2.0/55.0)*k5;
		a += system.output(y);
		//if(i%sys.sample==0) { a *= (1.0/sys.sample); r.paste(a,(i/sys.sample)*sys.states); a.clear(); }
		if(i%sys.sample==0) { a = system.output(y); r.paste(a,(i/sys.sample)*sys.states); }
		system.last(y);
	}
	
	return r;
}

void Solver::solve(Output& o)
{
	vector y(system.initial());
	vector r(sys.states*sys.steps,ROW);
	vector k0(y.dim(),ROW);
	vector k1(y.dim(),ROW);
	vector k2(y.dim(),ROW);
	vector k3(y.dim(),ROW);
	vector k4(y.dim(),ROW);
	vector k5(y.dim(),ROW);
	vector a(sys.states,ROW);
	double t = 0.0;

	int prog = 0;
	tty("[");
	for(int i=0;i<steps;i++,t+=sys.istep)
	{
		k0 = sys.istep * system.righthand(y,t);
		k1 = sys.istep * system.righthand(y + (1.0/4.0) * k0,t + (1.0/4.0) * sys.istep);
		k2 = sys.istep * system.righthand(y + (3.0/32.0) * k0 + (9.0/32.0) * k1,t + (3.0/8.0) * sys.istep);
		k3 = sys.istep * system.righthand(y + (1932.0/2197.0) * k0 - (7200.0/2197.0) * k1 + (7296.0/2197.0) * k2,t + (12.0/13.0) * sys.istep);
		k4 = sys.istep * system.righthand(y + (439.0/216.0) * k0 - 8.0 * k1 + (3680.0/513.0) * k2 - (845.0/4104.0) * k3,t + sys.istep);
		k5 = sys.istep * system.righthand(y - (8.0/27.0) * k0 + 2.0 * k1 - (3544.0/2565.0) * k2 + (1859.0/4104.0) * k3 - (11.0/40.0) * k4,t + (1.0/2.0) * sys.istep);
		y += (16.0/135.0)*k0 + (6656.0/12825.0)*k2 + (28561.0/56430.0)*k3 - (9.0/50.0)*k4 + (2.0/55.0)*k5;
		a += system.output(y);
		//if(i%sys.sample==0) { o.set((i/sys.sample),y.copy(0,sys.states),(1.0/sys.sample)*a/*+drift.local(i/sys.sample)*drift.getpara()*/); a.clear(); }
		if(i%sys.sample==0) { o.set((i/sys.sample),y.copy(0,sys.states),system.output(y)+theta(sys.drift_order)*drift.local(i/sys.sample)*drift.getpara()); a.clear(); }
		if(9.5*t/(sys.steps*sys.step)>prog) { prog++; tty("|"); }
		system.last(y);
	}
	say("]");

	return;
}
///*

#endif
