///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (system.hh)
///*

///guard
#ifndef HH_SYSTEM
#define HH_SYSTEM
///*

///header
/* class name:	System
 * 
 * description:	super class combing dynamic and forward system
 * 
 * author:		Christian Himpe
 * 
 * notes:		...
 * 
 * todo:        ...
 * 
 * version:		1.0
 */
///*

///includes
#include <vector>
#include "structs.hh"
#include "vector.hh"
#include "matrix.hh"
#include "structs.hh"
#include "dynlin.hh"
#include "dyneeg.hh"
//#include "dyneee.hh"
#include "forhem.hh"
#include "foreeg.hh"
#include "forunt.hh"
#include "feed.hh"
///*

///declarations
typedef std::vector<Forbase*> Forbasearray;
///*

///definitions
class System
{
	private:
		const syspara sys;												//General Parameter Structure
		Feed& feed;														//Data and Input Provider
		Dynbase* dynamic;												//Dynamic System
		Forbasearray forward;											//Array of Forward Systems
		int dim;														//Total Dimension of Dynamic and Forward Systems
		int ddim;														//Dynamic System Dimension
		int fdim;														//Forward System Dimension
		int paras;														//Total Number of Parameters
		int dparas;														//Dynamic System Parameters
		int fparas;														//Forward System Parameters 
	public:	
		System(const syspara& s,Feed& e,void* d=0,void* f=0,bool b=0);  //Constructor
		~System() { ; };												//Destructor
		vector initial() const;											//Initial Values
		vector righthand(const vector& v,double t);						//Righthandside
		vector output(const vector& v) const;							//Compute Output
		void   reset(const vector& v);									//Reset Parameters
		void   last(const vector& v);									//Set Previous Result as Backstep					
		vector getpara(bool t=0) const;									//Get Parameter Vector
		matrix getcova() const;											//Get Parameter Covariance Matrix
		int    getdim() const { return dim; };							//Get Total Dimension
		int    getparas() const { return paras; };						//Get Number of Parameters
		bool   isstable() const { return dynamic->isstable(); }			//Is Dynamic System Stable
};
///*

///implementation
System::System(const syspara& s,Feed& e,void* d,void* f,bool b) : sys(s), feed(e)
{
	char dt = 'X';
	char ft = 'X';
	switch(sys.dtype)
	{
		case DYNLIN: dynamic = new Dynlin(s,d); dt = 'm'; break;
		case DYNEEG: dynamic = new Dyneeg(s,d); dt = 'e'; break;
		//case DYNEEE: dynamic = new Dyneee(s,d); dt = 'E'; break;
		default: assert(0); break;
	}
	switch(sys.ftype)
	{
		case FORHEM: for(int i=0;i<sys.states;i++) { forward.push_back(new Forhem(s,i,f)); } ft = 'm'; break;
		case FOREEG: for(int i=0;i<sys.states;i++) { forward.push_back(new Foreeg(s,i,f)); } ft = 'E'; break;
		case FORUNT: for(int i=0;i<sys.states;i++) { forward.push_back(new Forunt(s,i,f)); } ft = 'e'; break;
		default: assert(0); break;
	}

	ddim = dynamic->getdim();
	fdim = forward[0]->getdim();
	dim = ddim+fdim*sys.states;
	dparas = dynamic->getparas();
	fparas = forward[0]->getparas();
	paras = dparas + sys.states*fparas;
	
	if(b==1) { tty("[    "); tty(dt); tty(ft); say("    ]"); }
}

vector System::initial() const
{
	vector r(dim,ROW);
	r.paste(dynamic->initial(),0);
	if(fdim!=0) { for(int i=0;i<sys.states;i++) { r.paste(forward[i]->initial(),ddim+i*fdim); } }
	return r;
}

vector System::righthand(const vector& v,double t)
{
	vector r(dim,ROW);
	r.paste(dynamic->righthand(v.copy(0,ddim),feed.jnput(t),t),0);
	if(fdim!=0) { for(int i=0;i<sys.states;i++) { r.paste(forward[i]->righthand(v(i),v.copy(ddim+i*fdim,fdim)),ddim+i*fdim); } }
	return r;
}

vector System::output(const vector& v) const
{
	vector r(sys.states,ROW);
	for(int i=0;i<sys.states;i++) { r(i,forward[i]->getoutput(v(i),v.copy(ddim+i*fdim,fdim))); }
	return r;
}

void System::reset(const vector& v)
{
	dynamic->reset(v.copy(0,dparas));
	if(fparas!=0) { for(int i=0;i<sys.states;i++) { forward[i]->reset(v.copy(dparas+i*fparas,fparas)); } }
}

void System::last(const vector& v)
{
	dynamic->last(v.copy(0,ddim));
}

vector System::getpara(bool t) const
{
	vector d(dynamic->getpara(t));
	vector r(d.dim()+sys.states*fparas,ROW);
	r.paste(d,0);
	if(fparas!=0) { for(int i=0;i<sys.states;i++) { r.paste(forward[i]->getpara(),d.dim()+i*fparas); } }
	return r;

}

matrix System::getcova() const
{
	matrix r(paras,paras);
	r.paste(dynamic->getcova(),0,0);
	if(fparas!=0) { for(int i=0;i<sys.states;i++) { r.paste(forward[i]->getcova(),dparas+i*fparas,dparas+i*fparas); } }
	return r;
}
///*

#endif
