///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (utils.hh)
///*
 
///guard
#ifndef HH_UTILS
#define HH_UTILS
///*

///includes
#include <stdlib.h>
#include <sys/resource.h>
#include <iostream>
#include <iomanip>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <omp.h>
///*

///declarations
typedef std::vector<int> intarray;
typedef std::vector<double> doublearray;
///*

///definitions
int threads = 1;

void   limit(uint gb);
void   maxthreads();
template<typename T>void tty(T c);
void   tty(double c);
template<typename T>void say(T c);
void   say(double c);
void   sys(const char* c);
void   ok();
void   fail();
int    count(const char* c);
const char* base(const char* p);
double sign(double x);
double theta(double x);
int    iround(double x);
double min(double x,double y);
double max(double x,double y);
double factorial(int x);
double gamma(double x);
double gauss(double mu=0.0,double sigma=1.0);
double loggauss(double mu=0.0,double sigma=1.0);
double cumnorm(double x);
double invcumnorm(double x);
double invcumchi2(double n,double x);
///*

///implementation
void limit(uint gb)
{
	rlimit c;
	getrlimit(RLIMIT_AS,&c);
	c.rlim_cur = c.rlim_max = long(1024*1024*1024)*long(gb);
	setrlimit(RLIMIT_AS,&c);
	tty("[   "); if(gb<9) { tty(" "); } tty(gb); say("GB   ]");
}

void maxthreads()
{
	threads = omp_get_num_procs();
	if(threads>1) { threads--; }
	if(threads>4) { threads--; }
	std::cout << "[    " << std::setw(2) << std::setfill(' ') << threads << "    ]" << std::endl;
}

template<typename T>void tty(T c) { std::cout << c << std::flush; }
void tty(double c) { std::cout.precision(10); std::cout << std::fixed << c << std::flush; }
template<typename T>void say(T c) { std::cout << c << std::endl; }
void say(double c) { std::cout.precision(10); std::cout << std::fixed << c << std::endl; }
void sys(const char* c) { std::cout << std::setw(68) << std::setfill('.') << std::left << c; }
void ok() { std::cout << "[    OK    ]" << std::endl; }
void fail() { std::cout << "[   FAIL   ]" << std::endl; }
int  count(const char* c, char s) { int r=0; for(int i=0;i<strlen(c);i++) { if(c[i]==s) { r++; } } return r; } 
const char* base(const char* p) { const char* r = strrchr(p,'/'); r = r?&(r[1]):p; return r; }
double  sign(double x) { return double((x>0.0)-(x<0.0)); }
double  theta(double x) { return double(x>=0.0); }
int  iround(double x) { return int(x+0.5); }
double min(double x,double y) { return (x<y)?x:y; }
double max(double x,double y) { return (y<x)?x:y; }

double factorial(int x)
{
	double r = 1;
	for(int i=2;i<=x;++i) { r *= double(i); }
	return r;
}

double gamma(double x) //Simple Stirling Approximation
{
	return sqrt(2.0*M_PI/x)*pow(x,x)/exp(x);
}

double gauss(double mu,double sigma) //Box-Mueller
{
	bool issigma = 0;
	static double svsigma = 0.0;
	double polar = 0.0;
	double rsquared = 0.0;
	double var1 = 0.0;
	double var2 = 0.0;
	
	if(!issigma)
	{
		do
		{
			var1 = 2.0 * (double(rand())/double(RAND_MAX) ) - 1.0;
			var2 = 2.0 * (double(rand())/double(RAND_MAX) ) - 1.0;
			rsquared = var1 * var1 + var2 * var2;
		}
		while(rsquared>=1.0 || rsquared==0.0);
		
		polar = sqrt(-2.0*log(rsquared)/rsquared);
		svsigma = var1 * polar;
		issigma = true;
		return (var2 * polar * sigma + mu);
	}
	else
	{
		issigma = false;
		return (svsigma * sigma + mu);
	}
}

double loggauss(double mu,double sigma)
{
	if(mu==0.0) { return exp(gauss(0,sigma)); }
	return exp(gauss(log(mu),sigma));
}

double cumnorm(double x)
{
	const double b1 =  0.319381530;
	const double b2 = -0.356563782;
	const double b3 =  1.781477937;
	const double b4 = -1.821255978;
	const double b5 =  1.330274429;
	const double p  =  0.2316419;
	const double c  =  0.39894228;
	double t = 0.0;
	double r = 0.0;

	if(x>=0.0)
	{
		t = 1.0 / ( 1.0 + p * x );
		r = (1.0 - c * exp( -x * x / 2.0 ) * t * ( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
	}
	else
	{
		t = 1.0 / ( 1.0 - p * x );
		r =      ( c * exp( -x * x / 2.0 ) * t * ( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
	}
	
	return r;
}

double invcumnorm(double x)
{
	const double s = 0.42;
	const double a0 = 2.50662823884;
	const double a1 = -18.61500062529;
	const double a2 = 41.39119773534;
	const double a3 = -25.44106049637;
	const double b1 = -8.47351093090;
	const double b2 = 23.08336743743;
	const double b3 = -21.06224101826;
	const double b4 = 3.13082909833;
	const double c0 = -2.78718931138;
	const double c1 = 4.85014127135;
	const double c2 = 4.85014127135;
	const double c3 = 2.32121276858;
	const double d1 = 3.54388924762;
	const double d2 = 1.63706781897;

	double q = x - 0.5;
	double r = 0.0;
	double p = 0.0;
	if(fabs(q)<=s) { r = q*q; return q*(((a3*r + a2)*r + a1)*r + a0)/((((b4*r + b3)*r + b2)*r + b1)*r + 1.0); }
	r = x;
	if(q>0.0) { r = 1.0 - x; }
	if(r>0.0)
	{
		r = sqrt(-log(r));
		p = (((c3*r + c2)*r + c1)*r + c0)/((d2*r + d1)*r + 1.0);
		if(q<0.0) { p = -p; }
		return p;
	}

	return 0.0;
}

double invcumchi2(double n,double x)
{
	const double a1 = -0.9911;
	const double a2 = -0.6763;
	const double b1 = 0.8055;
	const double b2 = -1.2451;
	double c1 = -log(2.0*(1.0-x));
	double c2 = -log(2.0*x);
	double z = 0.0;
	if(x>=0.5) { z = (-b1+sqrt(b1*b1-4.0*a1*c1))/(2.0*a1); }
	else       { z = (-b2-sqrt(b2*b2-4.0*a2*c2))/(2.0*a2); }
	return (z+sqrt(n))*(z+sqrt(n));
}

//void prefetch(void* hint) { __asm__ __volatile__ ("prefetch %%0": :"r"(hint) ); }
///*

#endif
