///license
//Christian Himpe 2010
//DCM - Dynamic Causal Modelling (vector.hh)
///*
 
///guard
#ifndef HH_VECTOR
#define HH_VECTOR
///*

///header
/* class name:	vector
 * 
 * description:	(dense) vector class
 * 
 * author:		Christian Himpe
 * 
 * notes:		...
 * 
 * todo:        ...
 * 
 * version:		1.0
 */
///*

///includes
#include <cmath>
#include <iostream>
#include <cassert>
#include <cstring>
#include <iomanip>
#include <cstdio>
#include <vector>
#include <deque>
#include <omp.h>
#include "utils.hh"
///*


///declarations
#define ROW false
#define COL true

class vector;
class matrix;
typedef std::deque<vector>  vectorque;
typedef std::vector<vector> vectorarray;
typedef std::vector<double> densevector;
///*

///definitions
class vector
{
	private:
		int size;
		bool rc;
		densevector dv;
	public:
		vector(int x,bool t,double b=0.0);								//Constructor
		~vector() { ; };												//Destructor
		double  operator()(int x) const  __attribute__((fastcall));		//vector component read
		void    operator()(int x,double d);								//vector component write
		vector  operator-() const;										//vector negation
		vector  operator+(const vector& v) const;						//vector addition
		vector  operator-(const vector& v) const;						//vector subtraction
		vector  operator*(double x) const;								//vector scalar multiplication
		double  operator*(const vector& v) const;						//vector dot product
		vector& operator+=(const vector& v);							//vector assignment addition
		vector& operator-=(const vector& v);							//vector assignment subtraction
		vector& operator*=(double x);									//vector assignment scalar multiplication
		vector& operator=(const vector& v);								//vector assignment
		
		int     dim() const { return size; };							//vector size
		double  length() const;											//vector length
		bool    type() const { return rc; };							//vector type (row or column)	
		vector  hadamard(const vector& v) const;						//vector hadamard product			
		vector  copy(int ro,int s) const;								//sub-vector (from to)
		void    paste(const vector& v,int r);							//copy vector (v to d into this)
		vector	transpose() const;											//transpose vector
		void    saxpy(const vector& v,double x);						//saxpy operation (this += x*v)
		double  max() const;											//find maximum vector entry
		double  min() const;											//find minimum vector entry
		vector  reciprocal() const;										//reciprocal value of each vector component
		vector  absolute() const;										//absolute value of each vector component
		void    clear(); 												//clear vector
		void    fill(double* d);										//fill vector with array
		void    print() const;											//print vector
		
		friend vector operator*(double x,const vector& v);				//friend vector left scalar multiplication
		friend vector operator*(const vector& v,const matrix& m);		//friend vector matrix multiplication
		friend matrix dyadic(const vector& v,const vector& w);			//friend vector dyadic product
		friend matrix diag(const vector& v);							//generate diagonal matrix from vector
};
///*

///implementation
vector::vector(int x,bool t,double b) : size(x),rc(t),dv(x,b) { ; }

double vector::operator()(int x) const
{
	assert(x<size);
	return dv[x]; 
}

void vector::operator()(int x,double d)
{
	assert(x<size);
	dv[x] = d;
	return;
}

vector vector::operator-() const
{
	vector r(*this);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for shared(r)
	for(int i=0;i<size;++i) { r.dv[i] = -dv[i]; }
	return r;
}

vector vector::operator+(const vector& v) const 
{
	assert(size==v.size);
	vector r(*this);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for shared(r)
	for(int i=0;i<size;++i) { r.dv[i] += v.dv[i]; }
	return r;
}

vector vector::operator-(const vector& v) const
{
	assert(size==v.size);
	vector r(*this);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for shared(r)
	for(int i=0;i<size;++i) { r.dv[i] -= v.dv[i]; }
	return r;
}

vector vector::operator*(double x) const
{
	vector r(*this);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for shared(r)
	for(int i=0;i<size;++i) { r.dv[i] *= x; }
	return r;
}

double vector::operator*(const vector& v) const
{
	assert(size==v.size);
	double r = 0.0;
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for reduction(+:r)
	for(int i=0;i<size;++i) { r += dv[i] * v.dv[i]; }
	return r;
}

vector& vector::operator+=(const vector& v)
{
	assert(size==v.size);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for 
	for(int i=0;i<size;++i) { dv[i] += v.dv[i]; }
	return *this;
}

vector& vector::operator-=(const vector& v)
{
	assert(size==v.size);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for 
	for(int i=0;i<size;++i) { dv[i] -= v.dv[i]; }
	return *this;
}

vector& vector::operator*=(double x)
{
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for 
	for(int i=0;i<size;++i) { dv[i] *= x; }
	return *this;
}

vector& vector::operator=(const vector& v)
{
	rc = v.rc;
	size = v.size;
	dv = v.dv;
	return *this;
}

double vector::length() const
{
	double r = 0.0;
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for reduction(+:r)
	for(int i=0;i<size;++i) { r += dv[i] * dv[i]; }
	return sqrt(r);
}

vector vector::hadamard(const vector& v) const
{
	assert(size==v.size);
	vector r(*this);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for shared(r)
	for(int i=0;i<size;++i) { r.dv[i] *= v.dv[i]; }
	return r;
}

vector vector::copy(int ro,int s) const
{
	assert(size>=ro+s);
	vector r(s,rc);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for shared(r) 
	for(int i=0;i<s;++i) { r.dv[i] = dv[ro+i]; }
	return r;
}

void vector::paste(const vector& v,int r)
{
	assert(size>=r+v.size);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for 
	for(int i=0;i<v.size;++i) { dv[r+i] = v.dv[i]; }
}

vector vector::transpose() const
{
	vector r(*this);
	r.rc = !rc;
	return r;
}

void vector::saxpy(const vector& v,double x)
{
	assert(size==v.size);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for 
	for(int i=0;i<size;++i) { dv[i] += v.dv[i]*x; }
}

double vector::max() const
{
	double r = dv[0];
	for(int i=1;i<size;++i) { r = ::max(r,dv[i]); }
	return r;
}

double vector::min() const
{
	double r = dv[0];
	for(int i=1;i<size;++i) { r = ::min(r,dv[i]); }
	return r;
}

vector vector::reciprocal() const
{
	vector r(size,rc);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for shared(r) 
	for(int i=0;i<size;++i) { r.dv[i] = 1.0/dv[i]; }
	return r;
}

vector vector::absolute() const
{
	vector r(size,rc);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for shared(r) 
	for(int i=0;i<size;++i) { r.dv[i] = fabs(dv[i]); }
	return r;
}

void vector::clear()
{
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for
	for(int i=0;i<size;++i) { dv[i] = 0.0; }
}

void vector::fill(double* d)
{
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for 
	for(int i=0;i<size;++i) { dv[i] = d[i]; }
}

void vector::print() const
{
	std::cout.precision(3);

	std::cout << "(";
	for(int i=0;i<size;i++)
	{
		std::cout << " ";
		std::cout << std::setfill(' ') << std::fixed << std::setw(6) << dv[i] << " "; 
	}
	std::cout << " )"; 
	if(rc==ROW) { std::cout << " T" << std::endl; }
	std::cout << std::endl; 
}

vector operator*(double x,const vector& v)
{
	vector r(v);
	//~ omp_set_num_threads(threads);
	//~ #pragma omp parallel for shared(r)
	for(int i=0;i<v.size;++i) { r.dv[i] *= x; }
	return r;
}
///*

#endif
